//------------------------------------------------------------------------------
//                   Copyright (c) 2003-2007 Codalogic Ltd.
//
// This file forms part of THE SUPPORTING SOFTWARE of Codalogic's LMX W3C
// Schema to C++ code generator product.  It may not be used independent of
// files legitimately generated and used by the LMX product.  See the LMX
// licence agreement for further restrictions on the use of this file.
//
// THIS FILE AND THE SOFTWARE CONTAINED HEREIN IS PROVIDED 'AS IS' AND COMES
// WITH NO WARRANTIES OF ANY KIND.
//
// CODALOGIC LTD SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY ANYONE
// OR ANYTHING DUE TO THE USE OF THIS FILE HOWEVER THEY MAY BE CAUSED.
//
// For more information on the LMX compiler, go to:
//
//     http://www.xml2cpp.com
//
// Also see:
//
//     http://www.codalogic.com
//
// Version: 3.6
//      - Added c_empty to allow for multiple empty elements.
//
// Version: 3.4
//      - Added ELMX_UNKNOWN_XSI_TYPE and ELMX_TYPE_NOT_SET
//        error codes.
//
// Version: 3.3
//      - Added QName, from_hex and from_base64 helper functions.
//
// Version: 3.2
//      - Added ELMX_VALUE_DOES_NOT_MATCH_FIXED_CONSTRAINT error constant.
//
// Version: 3.0
//      - Used std:: versions of a number of C utility functions.
//      - Added definition of t_class_identity.
//
// Version: 3.0 (Beta)
//      - Moved xisxxxx() functions to this header file from lmxtypes.h
//      - ELMX_UNDEFINED_ERROR error code added.
//
// Version: 2.11.3
//      - Added virtual destructor to c_lmx_exception to remove GCC 4.x
//        warnings.
//
// Version: 2.11
//      - No changes
//
// Version: 2.10
//      - Unicode handling extended beyond the Basic Multi-lingual Plane
//        to full Unicode range (0x-0x10ffff).
//
// Version: 2.9.2
//      - Added switches for MS VC 2005
//
// Version: 2.9
//      - No changes
//
// Version: 2.8
//      - No changes
//
// Version: 2.7
//      - Moved inittowidestring as an inline function into this file allow
//        control of wide vs. narrow string usage for customers without
//        supporting software source code.
//
// Version: 2.6
//
// Version: 2.5
//
//------------------------------------------------------------------------------


#ifndef LMXINTERNALS_H
#define LMXINTERNALS_H

#if defined( _MSC_VER ) && _MSC_VER > 1000
#pragma once
#pragma warning(disable:4786)
#endif

#if defined( _MSC_VER ) && _MSC_VER >= 1400
#pragma warning (push)
#pragma warning (disable:4996)  // Disable VC8 'safe' warnings
#endif

#ifndef LMXUSER_H
#error lmxinternals.h should now be included via lmxuser.h in order to setup declspec
#endif

#if defined( __BORLANDC__ )
    #pragma warn -8004
    #pragma warn -8012
#endif

#include <cctype>
#include <string>
#include <vector>

namespace lmx {

//------------------------------------------------------------------------------
//                             xs:all ordering constants
//------------------------------------------------------------------------------

const size_t k_all_order_not_present = ~0;

//------------------------------------------------------------------------------
//                             Error codes
//------------------------------------------------------------------------------

enum elmx_error_type { EET_FATAL, EET_ERROR, EET_WARNING, EET_NOTE, EET_NONE };

enum elmx_error {
    ELMX_OK,
    ELMX_NO_FILE,
    ELMX_ALL_CARDINALITY_VIOLATION,
    ELMX_ATTRIBUTE_ALREADY_READ,
    ELMX_ATTRIBUTES_IN_ANONYMOUS_COMPOSITOR,
    ELMX_BADLY_FORMED,
    ELMX_BAD_CHAR_ENCODING,
    ELMX_BAD_CHOICE,
    ELMX_BAD_COMMENT,
    ELMX_BAD_DTD,
    ELMX_BAD_ELEMENT_END,
    ELMX_BAD_END_OF_START_TAG,
    ELMX_BAD_END_TAG,
    ELMX_BAD_XML_DECL,
    ELMX_BAD_XML_VERSION,
    ELMX_CHAR_ENCODING_MISMATCH,
    ELMX_ELEMENT_NOT_FOUND,
    ELMX_EXTERNAL_ENTITY,
    ELMX_FRACTION_DIGITS_EXCEEDED,
    ELMX_LENGTH_TOO_LONG,
    ELMX_LENGTH_TOO_SHORT,
    ELMX_MANDATORY_ELEMENT_MISSING,
    ELMX_NO_PATTERN_MATCHED,
    ELMX_NOT_WELL_FORMED,
    ELMX_RECURSIVE_ENTITY_DEFINITION,
    ELMX_REQUIRED_ATTRIBUTES_MISSING,
    ELMX_TOO_MANY_ITEMS,
    ELMX_TOTAL_DIGITS_EXCEEDED,
    ELMX_UNABLE_TO_READ_ATTRIBUTE_VALUE,
    ELMX_UNABLE_TO_READ_ELEMENT_VALUE,
    ELMX_UNEXPECTED_ALL_EVENT,
    ELMX_UNEXPECTED_ELEMENT_EVENT,
    ELMX_UNEXPECTED_EMPTY_ELEMENT,
    ELMX_UNEXPECTED_ENTITY,
    ELMX_UNEXPECTED_EOM,
    ELMX_VALUE_BAD_FORMAT,
    ELMX_VALUE_EXCEEDS_MAX,
    ELMX_VALUE_EXCEEDS_MIN,
    ELMX_VALUE_NOT_ENUMERATED,
    //------------------------
    // Added in V2.1
    ELMX_UNEXPECTED_ATTRIBUTE,
    ELMX_NILLED_ELEMENT_NOT_EMPTY,
    //------------------------
    // Added in V2.7
    ELMX_ENTITY_NOT_FOUND,
    //------------------------
    // Added in V3.0
    ELMX_UNDEFINED_ERROR,   // Used as an initialization value that can be later overwritten with a valid error code
    //------------------------
    // Added in V3.2
    ELMX_VALUE_DOES_NOT_MATCH_FIXED_CONSTRAINT,
    //------------------------
    // Added in V3.4
    ELMX_UNKNOWN_XSI_TYPE,
    ELMX_TYPE_NOT_SET,
    //------------------------
    ELMX_LAST_ERROR
};

enum elxp_xml_event
{
    // General events
    EXE_EOM,
    EXE_UNKNOWN,
    EXE_FATAL_ERROR,

    // XML events
    EXE_BADLY_FORMED,
    EXE_START_TAG_END,
    EXE_CDATA,
    EXE_ELEMENT_END,
    EXE_TOKEN,

    EXE_XSI_TYPE,
    EXE_XSI_NIL,
    EXE_XSI_SCHEMALOCATION,
    EXE_XSI_NONAMESPACESCHEMALOCATION,

    EXE_LAST_XML_EVENT
};

//------------------------------------------------------------------------------
//                    Utility functions for initialising constants
//------------------------------------------------------------------------------

inline tlmx_unicode_string inittowstring( const char * ap_string )
{
    tlmx_unicode_string l_string;
    return vtostring( l_string, ap_string );
}

typedef char tc_utf8_buf[6+1];

// V3.3
inline tc_qname make_qname( const std::string &ar_qname, const std::string &ar_namespace, const std::string &ar_local_name )
{
    tc_qname l_qname;
    make_qname( &l_qname, ar_qname, ar_namespace, ar_local_name );
    return l_qname;
}

inline tc_binary from_hex( const std::string &ar_in )
{
    tc_binary l_binary;
    l_binary.from_hex( ar_in );
    return l_binary;
}

inline tc_binary from_base64( const std::string &ar_in )
{
    tc_binary l_binary;
    l_binary.from_base64( ar_in );
    return l_binary;
}


//------------------------------------------------------------------------------
//                               LMX exception classes
// These are designed to be non-throwing.  Interogate c_xml_reader or
// c_xml_writer within the catch block to find out more information.
//------------------------------------------------------------------------------

LMX_PDECL const char *get_error_description( elmx_error a_error_code );

class LMX_PDECL c_lmx_exception
{
private:
    elmx_error error_code;

public:
    c_lmx_exception( elmx_error a_error_code ) : error_code( a_error_code ) {}
    virtual ~c_lmx_exception() {}   // v2.11.3
    elmx_error get_error() const { return error_code; }
    virtual const char *what() const throw()
    {
        return get_error_description( error_code );
    }
};

class LMX_PDECL c_lmx_reader_exception : public c_lmx_exception
{
private:
    int line;

public:
    c_lmx_reader_exception( elmx_error a_error_code, int a_line )
            : c_lmx_exception( a_error_code ), line( a_line )
    {}
    int get_line() const { return line; }
};

//------------------------------------------------------------------------------
//                 Reader and writer class forward references
//------------------------------------------------------------------------------

typedef int t_xml_event;
typedef const char * t_class_identity;
struct s_event_map;
struct s_ns_map;    // Added in V2.9.2
class c_xml_reader;
class c_xml_writer;

//------------------------------------------------------------------------------
//                            Convenience typedefs
//------------------------------------------------------------------------------

class c_any_info
{
private:
    std::string o_namespace;
    std::string name;
    std::string value;

public:
    c_any_info()
    {
    }
    // For writing during unmarshaling
    c_any_info( const std::string &ar_namespace,
                            const std::string &ar_name,
                            const std::string &ar_value ) :
        o_namespace( ar_namespace ),
        name( ar_name ),
        value( ar_value )
    {
    }
    // For user writing xs:anyAttribute
    c_any_info( const std::string &ar_name,
                            const std::string &ar_value ) :
        name( ar_name ),
        value( ar_value )
    {
    }
    // For user writing xs:any
    c_any_info( const std::string &ar_value ) :
        value( ar_value )
    {
    }
    const std::string & get_namespace() const { return o_namespace; }
    const std::string & get_name() const { return name; }
    const std::string & get_value() const { return value; }
    void set_namespace( const std::string &ar_namespace ) { o_namespace = ar_namespace; }
    void set_name( const std::string &ar_name ) { name = ar_name; }
    void set_value( const std::string &ar_value ) { value = ar_value; }
};

typedef std::vector< c_any_info * > t_any_info_vector;

class c_empty
{
};

//------------------------------------------------------------------------------
//                   Signed char safe versions of isxxxx functions
//------------------------------------------------------------------------------

inline bool xisalpha( int x ) { return x > 0 && LMXSTDCRT::isalpha( x ); }
inline bool xisspace( int x ) { return x > 0 && LMXSTDCRT::isspace( x ); }
inline bool xisalnum( int x ) { return x > 0 && LMXSTDCRT::isalnum( x ); }
inline bool xisdigit( int x ) { return x > 0 && LMXSTDCRT::isdigit( x ); }
inline bool xisxdigit( int x ) { return x > 0 && LMXSTDCRT::isxdigit( x ); }
inline bool xisprint( int x ) { return x > 0 && LMXSTDCRT::isprint( x ); }


//------------------------------------------------------------------------------
//                         Compiler STL support fixes
//------------------------------------------------------------------------------

inline int stringncompare( const std::string &ar_lhs, const char *ap_rhs,
                        size_t a_first, size_t a_num )
{
#if defined( __GNUC__ ) && __GNUC__ == 2
    size_t l_len = strlen( ap_rhs );
    if( a_num > l_len ) a_num = l_len;
    return strncmp( ar_lhs.c_str() + a_first, ap_rhs, a_num );
#else
    return ar_lhs.compare( a_first, a_num, ap_rhs );
#endif
}
} // End of namespace lmx

#if defined( _MSC_VER ) && _MSC_VER >= 1400
#pragma warning (pop)  // Restore VC8 'safe' warnings
#endif

#endif // LMXINTERNALS_H


//------------------------------------------------------------------------------
//                  Copyright (c) 2003-2007 Codalogic Ltd.
//------------------------------------------------------------------------------

