//------------------------------------------------------------------------------
//                 Copyright (c) 2003-2007 Codalogic Ltd.
//
// This file forms part of THE SUPPORTING SOFTWARE of Codalogic's LMX W3C
// Schema to C++ code generator product.  It may not be used independent of
// files legitimately generated and used by the LMX product.  See the LMX
// licence agreement for further restrictions on the use of this file.
//
// THIS FILE AND THE SOFTWARE CONTAINED HEREIN IS PROVIDED 'AS IS' AND COMES
// WITH NO WARRANTIES OF ANY KIND.
//
// CODALOGIC LTD SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY ANYONE
// OR ANYTHING DUE TO THE USE OF THIS FILE HOWEVER THEY MAY BE CAUSED.
//
// For more information on the LMX compiler, go to:
//
//     http://www.xml2cpp.com
//
// Also see:
//
//     http://www.codalogic.com
//
// Version: 3.7
//      - Added features to facilitate working with SOAP messages, including:
//          -- Ability to read a named attribute.
//          -- Ability for writer to re-enable output of namespace maps.
//
// Version: 3.6.1
//      - Fixed querying for non-namespace.
//
// Version: 3.6
//      - Fixed some cases where entity expansion failures were not acted on
//        when using LMX_XML_STRICT.
//
// Version: 3.5
//      - Fixed handling of xsi-types with no namespace.
//
// Version: 3.4
//      - Added ELMX_UNKNOWN_XSI_TYPE and ELMX_TYPE_NOT_SET
//        error codes.
//
// Version: 3.3
//      - Added QName helper functions.
//      - Made LMX_XML_STRICT cause even stricter XML parsing.
//
// Version: 3.2.4
//      - Added better support for QName types.
//
// Version: 3.1
//      - Modified c_read_file::read() to read data a block at a time.  This
//        significantly improves performance under VS2005.
//      - Added ability for user to set schema locations when writing XML.
//
// Version: 3.0
//      - Used std:: versions of a number of C utility functions.
//      - Added code for expected_class_identity to c_xml_writer.
//
// Version: 3.0 (Beta)
//      - Parser completely refactored.
//
// Version: 2.11
//      - Removed use of global_event_map and added functions to support this.
//
// Version: 2.10.3
//      - Fix for rewind() operation
//
// Version: 2.10
//      - Unicode handling extended beyond the Basic Multi-lingual Plane
//        to full Unicode range (0x-0x10ffff).
//
// Version: 2.9.3
//      - Fixed C4251 warning for c_read_file under some build conditions.
//      - c_read_file can now be constructed from a FILE * handle to allow for
//        various read-modify-write scenarios, perhaps with file locking etc.
//
// Version: 2.9.2
//      - Fixed parsing of CDATA.
//      - Added switches for MS VC 2005
//
// Version: 2.9
//      - Moved float_zero() function into lmxparse.cpp to reduce chance of
//        divide by zero error.
//
// Version: 2.8
//      - Add ability to read from files with wchar_t specified names when using
//        MSVC++.
//      - Added definition of LMX_PDECL if it is not already defined.  During
//        normal compilation this is defined in lmxuser.h, but the extra
//        definition helps MS intellisense work correctly.
//
// Version: 2.7
//      - Some methods in c_xml changed from private to protected (for DTD parsing).
//      - Character entities expanded immediately in entity def, rather than entity
//        text used.
//
// Version: 2.6
//      - Modified constructor initialisation order to remove GCC -Wall
//        warnings.
//      - Added some defaults to switch statements to address compiling
//        with -Wall.
//      - Fixed problem with single quote appearing in attribute values
//        when parsing xs:any elements.
//      - Fixed problem with comments that contained no space after the
//        <!-- sequence and contained a / characters
//        (e.g. <!--My comment with / char -->)
//
// Version: 2.5.1
//      - Removed variable that became redundant in p_has_must_understand().
//      - Changed type of l_i from int to size_t to avoid -Wall
//        signed/unsigned comparison warning.
//
// Acknowledgements:
//      Borland additions kindly provided by Jeff Hay of HYTEC, Inc.,
//      http://www.hytecinc.com.  Find 'BORLANDC' to see additions.
//------------------------------------------------------------------------------


#ifndef LMXPARSE_H
#define LMXPARSE_H

#if defined( _MSC_VER ) && _MSC_VER > 1000
#pragma once
#pragma warning (disable:4786)  // Disable truncated debug info warning
#endif

#if defined( _MSC_VER ) && _MSC_VER >= 1400
#pragma warning (push)
#pragma warning (disable:4996)  // Disable VC8 'safe' warnings
#endif

#if defined( __BORLANDC__ )
#pragma warn -8027 // Disbale "not expanded inline" warnings
#endif

#include <string>
#include <vector>
#include <memory>
#include <cassert>
#include <iostream>
#include <fstream>  // For the marshal() template function
#include <sstream>  // For the marshal() template function

#include "lmxuser.h"

// To help MS intellisense
#if ! defined( LMX_PDECL )
#define LMX_PDECL
#endif

namespace lmx {

#if LMX_WANT_DLL == 0 && ! defined( LMX_MAKE_DLL ) && (defined( _MSC_VER ) || defined( __BORLANDC__ ))
    // USER: Try to warn user if they have not setup DLL imports correctly.
    // You may remove this section if you are happy with your project's linking.
    namespace NOTE_TO_USER_____
    {
        extern const int set_hash_define_LMX_WANT_DLL_to_1_to_use_DLL_versions_of_LMX_libraries___See_comments_near_LMX_WANT_DLL_in_lmxuser_h_file;
        const int user_note = set_hash_define_LMX_WANT_DLL_to_1_to_use_DLL_versions_of_LMX_libraries___See_comments_near_LMX_WANT_DLL_in_lmxuser_h_file;
    }
#endif

char *xstrdup( const char *);

//------------------------------------------------------------------------------
//                                 Error classes
//------------------------------------------------------------------------------

class c_get_as_utf8;    // Provides information on line numbers etc.
class c_getter_error_binder;

class LMX_PDECL c_error
{
protected:
    const c_get_as_utf8 *p_getter;
    int fatals;     // Treat fatals as pure errors also.  i.e. inc errors when fatals inced
    int errors;
    int warnings;
    int notes;      // Things the compiler does not support - could be switched off
    elmx_error_type last_error_type;
    elmx_error last_error_code;
    int last_error_line;

public:
    c_error() : p_getter(LMXNULL), fatals(0), errors(0), warnings(0), notes(0),
                last_error_type( EET_NONE ), last_error_code( ELMX_OK ), last_error_line(-1)
    {}
    // To ensure proper registering and unregistering of the getter, binding must be done
    // using an instance of the c_getter_error_binder class.
    const c_get_as_utf8 * register_getter( c_getter_error_binder *ap_binder );
    void unregister_getter( c_getter_error_binder *ap_binder );

    virtual void report_error( elmx_error_type a_type,
                     elmx_error a_code,
                     const char *ap_file,
                     int a_line,
                     const char *ap_msg ) {}

    // For no file, set ap_file to LMXNULL.  For no line, set a_line to -1.
    void error( elmx_error_type a_type,
            elmx_error a_code,
            const char *ap_file,
            int a_line,
            const char *ap_msg );
    // If p_getter is specified, c_error will ask file file and line information
    void error( elmx_error_type a_type,
            elmx_error a_code,
            const char *ap_msg );

    bool is_errored() { return (errors + fatals) ? true : false; }
    elmx_error get_last_error() { return last_error_code; }
    elmx_error_type get_last_error_type() { return last_error_type; }
    bool get_error_info( elmx_error_type *ap_type, elmx_error *ap_code, int *ap_line );
    bool is_fatal_errored() { return fatals != 0; }
    virtual ~c_error() {}
};

//------------------------------------------------------------------------------
//                            c_getter_error_binder
//               Implements binding between c_error and getter.
//   By making this the only way a getter can register with the error class
//       we ensure that the binding is always done and undone correctly.
//------------------------------------------------------------------------------

class LMX_PDECL c_getter_error_binder
{
private:
    bool is_bound;
    c_error &r_error;
    const c_get_as_utf8 *p_new_getter;
    const c_get_as_utf8 *p_old_getter;

public:
    c_getter_error_binder( c_error &ar_error );
    void bind( const c_get_as_utf8 *ap_getter );
    const c_get_as_utf8 *get_new_getter() const { return p_new_getter; }
    const c_get_as_utf8 *get_old_getter() const { return p_old_getter; }
    ~c_getter_error_binder();
};

enum elmx_whitespace { EXWS_PRESERVE, EXWS_REPLACE, EXWS_COLLAPSE };
enum elxp_attr_elem { EXNT_ATTR, EXNT_ELEM };
enum elxp_xml_known_namespaces {
    EXKN_UNKNOWN, EXKN_NONE, EXKN_XML, EXKN_XS, EXKN_XMLNS, EXKN_XSI, EXKN_TKWTYPES,
    EXKN_SOAP_ENV, EXKN_SOAP_ENC, EXKN_DYNAMIC_START = 1000 };

//------------------------------------------------------------------------------
//                      Unicode conversion functions
//------------------------------------------------------------------------------

LMX_PDECL bool utf8_to_unicode_char( tc_utf8_buf ac_utf8, t_unicode_char *ap_unicode );
LMX_PDECL char *unicode_char_to_utf8( t_unicode_char a_c, tc_utf8_buf ac_utf8 );
LMX_PDECL bool is_valid_utf8_char_sequence( const char ac_utf8[] );     // V3.2.4 - Added

//------------------------------------------------------------------------------
//                       Classes for low-level input
//                    Either memory mapped or File mapped
//------------------------------------------------------------------------------

class LMX_PDECL c_read
{
public:
    enum { k_eof = -1 };    // Can't be '\0' because UTF-16 and similar encodings contain 0

    c_read() {}

    virtual bool is_open() const { return true; }
    virtual int  read() = 0;
    virtual const char *get_source_id() { return LMXNULL; }
    virtual void rewind() = 0;

    virtual ~c_read() {}
};

class LMX_PDECL c_read_memory : public c_read
{
private:
    const char *p_message;
    const char *p_message_ptr;
    const char *p_message_end;

    c_read_memory() {}

public:
    c_read_memory( const char *ap_message, size_t a_message_length )
    {
        p_message_ptr = p_message = ap_message;
        p_message_end = ap_message + a_message_length;
    }
    ~c_read_memory()
    {
    }
    bool is_open() const
    {
        return true;
    }
    int read()
    {
        if( p_message_ptr >= p_message_end )
            return c_read::k_eof;

        return static_cast<unsigned char>(*p_message_ptr++);
    }
    void rewind()
    {
        p_message_ptr = p_message;
    }
};

class LMX_PDECL c_read_string : public c_read_memory
{
public:
    c_read_string( const std::string &ar_string ) : c_read_memory( ar_string.data(), ar_string.size() )
    {}
};

class c_read_file : public c_read
{
private:
    std::string file;
    // We could use ifstream.  However, the functionality provided by ifstream
    // -is way more than we need, and for the bits we want ultimately ends up calling
    // -the same functions as this.  Hence this should compile faster and run
    // -faster.
    LMXSTDCRT::FILE *h_input;
    bool is_file_to_be_closed_on_delete;
    char c_buf[512];
    size_t n_chars;
    size_t char_index;

private:
    c_read_file() {}

public:
    LMX_PDECL c_read_file( const char *ap_file )
        : file( ap_file ), is_file_to_be_closed_on_delete( true ), n_chars( 0 ), char_index( 0 )
    {
        h_input = LMXSTDCRT::fopen( ap_file, "rb" );
    }
#ifdef _MSC_VER
    LMX_PDECL c_read_file( const wchar_t *ap_file );
#endif
    LMX_PDECL c_read_file( LMXSTDCRT::FILE *ah_file )
        : file( "" ), is_file_to_be_closed_on_delete( true ), n_chars( 0 ), char_index( 0 )
    {
        h_input = ah_file;
    }
    LMX_PDECL void leave_file_open_on_delete( bool a_leave_open = true )
    {
        is_file_to_be_closed_on_delete = ! a_leave_open;
    }
    LMX_PDECL bool is_open() const
    {
        return (h_input)?true: false;
    }
    LMX_PDECL int read()
    {
        if( char_index >= n_chars ) // v3.1
            {
            if( ! h_input )
                return c_read::k_eof;
            n_chars = LMXSTDCRT::fread( c_buf, 1, sizeof(c_buf), h_input );
            if( n_chars == 0 )
                return c_read::k_eof;
            char_index = 0;
            }
        return static_cast<unsigned char>(c_buf[char_index++]);
    }
    LMX_PDECL void rewind()
    {
        LMXSTDCRT::fseek( h_input, 0, SEEK_SET );
    }
    LMX_PDECL const char * get_source_id()
    {
        return file.c_str();
    }
    LMX_PDECL void close()
    {
        if( h_input ) LMXSTDCRT::fclose( h_input );
        h_input = LMXNULL;
    }
    LMX_PDECL ~c_read_file()
    {
        if( is_file_to_be_closed_on_delete && h_input )
            LMXSTDCRT::fclose( h_input );
    }
};

//------------------------------------------------------------------------------
//                           c_unget_buffer
//           Allows multi character push back (min 2 required)
//------------------------------------------------------------------------------

class LMX_PDECL c_unget_buffer
{
private:
    enum { size = 3 };
    int c_buffer[size];
    size_t free_index; // index of next empty slot.  == 0 if empty

public:
    c_unget_buffer() : free_index( 0 ) {}
    void clear() { free_index = 0; }
    inline bool empty() const { return free_index == 0; }
    inline int get()
    {
        lmx_assert( free_index > 0 );
        return c_buffer[--free_index];
    }
    inline void unget( int a_c )
    {
        lmx_assert( free_index < size );
        c_buffer[free_index++] = a_c;
    }
};

//------------------------------------------------------------------------------
//                           class c_entity_name_reader
//              A base utility class that allows the name of entities to be
//                read from various input sources such as c_get_for_xml,
//                   other entities and the stored attribute data.
//------------------------------------------------------------------------------

class c_entity_name_reader
{
private:
    bool o_is_replacement_processed;

public:
    c_entity_name_reader() : o_is_replacement_processed( false ) {}
    virtual ~c_entity_name_reader() {}  // Needs virtual destructor in base

    virtual int get() = 0;      // Returns c_read::k_eof for end of text
    virtual void unget( int a_c ) = 0;

    void is_replacement_processed( bool a_is_processed ) { o_is_replacement_processed = a_is_processed; }
    bool is_replacement_processed() const { return o_is_replacement_processed; }
};

//------------------------------------------------------------------------------
//                   Data Repository (Namespaces and enitites)
//------------------------------------------------------------------------------

template <class T>
struct c_rep_ops
{
private:
    T data;
public:
    typedef T Tconst;
    c_rep_ops( Tconst a ) : data( a ) {}
    T get() { return data; }
    static bool eq( Tconst a, Tconst b ) { return a == b; }
    static void release( T a ) {}
    ~c_rep_ops() {}
};

template<> struct c_rep_ops< char * >
{
private:
    std::auto_ptr< char > p_data;
public:
    typedef const char * Tconst;
    c_rep_ops( Tconst a ) : p_data( xstrdup( a ) ) {}
    char * get() { return p_data.get(); }
    static bool eq( Tconst a, Tconst b ) { return strcmp( a, b ) == 0; }
    static void release( char * a ) { delete a; }
    ~c_rep_ops() { p_data.release(); }  // This object only exists while adding the data to the repository
};

template<> struct c_rep_ops< const char * >
{
private:
    std::auto_ptr< char > p_data;   // Must be < char > (not < const char >) for VC6's benefit
public:
    typedef const char * Tconst;
    c_rep_ops( Tconst a ) : p_data( xstrdup( a ) ) {}
    const char * get() { return p_data.get(); }
    static bool eq( Tconst a, Tconst b ) { return strcmp( a, b ) == 0; }
    static void release( const char * a ) { delete const_cast<char *>( a ); }   // The cast is for VC6's benefit!
    ~c_rep_ops() { p_data.release(); }  // This object only exists while adding the data to the repository
};

// This repository is expected to only have a few members, hence is likely to
// -be faster than a std::map and generate less code.
template <class T, class Tops = c_rep_ops<T> >
class c_repository
{
protected:
    typedef std::pair< char *, T > t_repository_item;
    typedef std::vector< t_repository_item > t_repository;
    t_repository repository;

public:
    void set( const char *ap_key, typename Tops::Tconst a_value )
    {
        // Scan database to see if key already exists
        typename t_repository::iterator li_db( repository.begin() );
        typename t_repository::iterator li_db_end( repository.end() );

        while( li_db != li_db_end )
            {
            if( strcmp( li_db->first, ap_key ) == 0 )
                {
                Tops::release( li_db->second );
                Tops l_ops( a_value );
                li_db->second = l_ops.get();
                return;
                }
            ++li_db;
            }

        // If key was not found, then insert value at end
        std::auto_ptr< char > l_key_copy( xstrdup( ap_key ) );
        Tops l_ops( a_value );
        repository.push_back( t_repository_item( l_key_copy.get(), l_ops.get() ) );
        l_key_copy.release();
    }
    bool get( const char *ap_key, T * ap_value ) const
    {
        typename t_repository::const_iterator li_db( repository.begin() );
        typename t_repository::const_iterator li_db_end( repository.end() );
        while( li_db != li_db_end )
            {
            if( strcmp( li_db->first, ap_key ) == 0 )
                {
                *ap_value = li_db->second;
                return true;
                }
            ++li_db;
            }
        return false;
    }
    const char * reverse_lookup( typename Tops::Tconst ar_value ) const
    {
        typename t_repository::const_iterator li_db( repository.begin() );
        typename t_repository::const_iterator li_db_end( repository.end() );
        while( li_db != li_db_end )
            {
            if( Tops::eq( li_db->second, ar_value ) )
                return li_db->first;
            ++li_db;
            }
        return LMXNULL;
    }
    bool is_defined( const char *ap_key ) const
    {
        typename t_repository::const_iterator li_db( repository.begin() );
        typename t_repository::const_iterator li_db_end( repository.end() );
        while( li_db != li_db_end )
            {
            if( strcmp( li_db->first, ap_key ) == 0 )
                return true;
            ++li_db;
            }
        return false;
    }
    virtual void clear()
    {
        typename t_repository::iterator li_db( repository.begin() );
        typename t_repository::iterator li_db_end( repository.end() );
        while( li_db != li_db_end )
            {
            delete li_db->first;
            Tops::release( li_db->second );
            ++li_db;
            }
        repository.resize( 0 );     // Hold on to any memory we've already allocated
    }
    virtual ~c_repository() { clear(); }
};

//------------------------------------------------------------------------------
//                            class c_get_as_utf8
//             Class for reading 'any' input and converting to UTF-8
//------------------------------------------------------------------------------

class c_get_as_utf8
{
public:
    enum e_encoding_mode { EXR_INIT, EXR_ISO8859_1, EXR_UTF8,
                            EXR_UTF16_BE, EXR_UTF16_LE,
                            EXR_UCS_2_BE, EXR_UCS_2_LE };

private:
    c_read &r_reader;
    c_error &r_error;
    c_getter_error_binder getter_error_binder;
    e_encoding_mode encoding_mode;
    char last_get;                  // Used for end of line fix-up
    c_unget_buffer unget_buffer;
    tc_utf8_buf c_utf8_expansion;   // Filled when ISO or UTF16 code is encountered
    char *p_utf8_expansion;         // LMXNULL if no utf8 expansion
    int line_num;

private:
    void p_init();
    void p_read_single_utf8_compound_char_sequence( int a_first_c, char ac_utf8[] );
    int p_read_utf8_compound_char_sequence( int a_first_c );
    int p_get();
    int p_read() { return r_reader.read(); }

    int p_resolve_initial_character_encoding( int a_initial_c );
    int p_map_unicode_to_utf8( t_unicode_char a_c );
    t_unicode_char p_convert_utf16_bytes( int l_b1, int l_b2 );

public:
    LMX_PDECL c_get_as_utf8( c_read &ar_reader, c_error &ar_error );
    LMX_PDECL virtual ~c_get_as_utf8();

    LMX_PDECL void rewind();

    LMX_PDECL bool set_encoding( const char *ap_encoding );

    LMX_PDECL int get();
    LMX_PDECL void unget( int a_returned ) { unget_buffer.unget( a_returned ); }

    LMX_PDECL const char *get_source_id() const { return r_reader.get_source_id(); }
    LMX_PDECL int get_line_num() const { return line_num; }

    // Error reporting is delegated to this class because it knows about line numbers, file names etc.
    LMX_PDECL c_error & get_error_handler() const { return r_error; }
    LMX_PDECL void error( elmx_error_type a_type, elmx_error a_code, const char *ap_format );
    LMX_PDECL void error( elmx_error_type a_type, elmx_error a_code, const std::string &ar_s );
};

//------------------------------------------------------------------------------
//                            class c_get_with_entities
//                    Adds the ability to read input from entities
//                            defined in an internal DTD.
//                         (c_dtd_active_entity is a helper)
//------------------------------------------------------------------------------

struct c_dtd_active_entity
{
    const char *p_current;
    const char *p_base;
    c_unget_buffer unget_buffer;

    c_dtd_active_entity( const char *ap_entity ) :
        p_current( ap_entity ),
        p_base( ap_entity )
    {}
    int get()
    {
        if( ! unget_buffer.empty() )
            return unget_buffer.get();
        lmx_assert( *p_current != '\0' );
        return *p_current++;
    }
    void unget( int a_c )
    {
        unget_buffer.unget( a_c );
    }
    bool empty() const
    {
        return unget_buffer.empty() && *p_current == '\0';
    }
};

class c_get_with_entities : public c_get_as_utf8
{
private:
    std::vector< c_dtd_active_entity > entity_input_stack;
    c_repository< const char * > general_entities;      // Of the form &entity;
    c_repository< const char * > parameter_entities;    // Of the form %entity; - Only valid in DTD but get() operation needs to be able to return contents

public:
    LMX_PDECL c_get_with_entities( c_read &ar_reader, c_error &ar_error );
    LMX_PDECL virtual ~c_get_with_entities();

    LMX_PDECL int get();
    LMX_PDECL void unget( int a_returned );

    LMX_PDECL void rewind()
    {
        general_entities.clear();
        parameter_entities.clear();
        c_get_as_utf8::rewind();
    }

    LMX_PDECL void add_general_entity( const char *ap_entity_tag, const char *ap_value )
    {
        general_entities.set( ap_entity_tag, ap_value );
    }
    LMX_PDECL bool get_general_entity( const char *ap_entity_tag, const char ** app_entity_value_out ) const
    {
        return general_entities.get( ap_entity_tag, app_entity_value_out );
    }
    LMX_PDECL void add_parameter_entity( const char *ap_entity_tag, const char *ap_value )
    {
        parameter_entities.set( ap_entity_tag, ap_value );
    }
    LMX_PDECL bool get_parameter_entity( const char *ap_entity_tag, const char ** app_entity_value_out ) const
    {
        return parameter_entities.get( ap_entity_tag, app_entity_value_out );
    }
    LMX_PDECL bool set_input_entity_value( const char *ap_input_text );

    LMX_PDECL size_t get_entity_stack_depth() const { return entity_input_stack.size(); }
};

//------------------------------------------------------------------------------
//                              class c_get_for_xml
//             Class for getting characters specific to XML and DTDs.
//               Mainly this adds white space handling and utility
//                     functions common to parsing XML and DTDs.
//------------------------------------------------------------------------------

class c_get_for_xml : public c_get_with_entities
{
private:
    std::string entity_tag;
    elmx_whitespace ws_mode;
    bool ws_proc_got_white;
    bool is_reading_xsd;    // White space handling of characters from entities is affected by this.

private:
    void p_init();

public:
    LMX_PDECL c_get_for_xml( c_read &ar_reader, c_error &ar_error, bool a_is_reading_xsd = true );
    LMX_PDECL virtual ~c_get_for_xml();

    LMX_PDECL void add_char_wsp( std::string *ap_value, int a_c, bool a_is_from_char_entity=false );    // Add char with white space processing
    LMX_PDECL void set_ws_mode( elmx_whitespace a_ws_mode );
    LMX_PDECL void reset_ws_mode() { set_ws_mode( ws_mode ); }

    LMX_PDECL int get_non_ws()  // Get next non-whitespace character
    {
        int l_c;
        while( (l_c = get()) <= ' ' && xisspace( l_c ) )    // Various line ends have been converted to 0x0a so all ws <= ASCII space.
            {}
        return l_c;
    }

    LMX_PDECL void rewind();

    LMX_PDECL bool insert_character_entity( c_entity_name_reader &ar_entity_tag_reader, std::string *ap_value );
    LMX_PDECL bool read_specified_string_from_input( const char * ap_input_string_to_be_read );
    LMX_PDECL bool get_simple_quoted_value( std::string *ap_value );
    LMX_PDECL bool read_xml_decl();
    LMX_PDECL bool skip_comment();
    LMX_PDECL bool skip_PI();
    LMX_PDECL bool move_to_start_tag_end( bool *ap_end_of_element = LMXNULL );
};

//------------------------------------------------------------------------------
//                              class c_dtd_util
//             Utilities common to reading internal and external DTDs.
//------------------------------------------------------------------------------

class c_dtd_util
{
private:
    c_get_for_xml &r_get;

public:
    LMX_PDECL c_dtd_util( c_get_for_xml &ar_get );

    LMX_PDECL c_get_for_xml & getter() { return r_get; }
    LMX_PDECL int get() { return r_get.get(); }
    LMX_PDECL void unget( int a_returned ) { r_get.unget( a_returned ); }

    LMX_PDECL bool get_entity_value( std::string *ap_value );
    LMX_PDECL bool get_quoted_entity_value( std::string *ap_value );
    LMX_PDECL bool select_parameter_entity();
};

//------------------------------------------------------------------------------
//                              class c_internal_dtd
//                Extracts user defined entities from an internal DTD.
//                  Definitions of elements, attributes and all other
//                             such items are ignored.
//------------------------------------------------------------------------------

class c_internal_dtd
{
private:
    c_dtd_util dtd_util;

private:
    void p_get_markupdecl_name( std::string *ap_name ); // Different to external DTD form (does not read external references)
    bool p_read_dtd_entity();
    bool p_skip_to_end_of_markupdecl();

public:
    LMX_PDECL c_internal_dtd( c_get_for_xml &ar_get );

    LMX_PDECL bool read_doctype_decl();
    LMX_PDECL void reset();
};

//------------------------------------------------------------------------------
//                          class c_enumerated_namespaces
//           Allows namespace names to be mapped to integer enumerations.
//             This allows namespaces to be more efficiently handled.
//------------------------------------------------------------------------------

class c_enumerated_namespaces
{
private:
    int                 next_enumeration;
    c_repository<int>   namespaces;

public:
    LMX_PDECL c_enumerated_namespaces();
    LMX_PDECL ~c_enumerated_namespaces() {}

    LMX_PDECL void clear();
    LMX_PDECL int find_namespace( const std::string &ar_namespace_uri ) const;
    LMX_PDECL const char * find_namespace_from_id( int a_namespace_id ) const;
    LMX_PDECL int add_namespace( const char *ap_namespace_uri );
    LMX_PDECL void add_namespace( const char *ap_namespace_uri, int a_value );
};

//------------------------------------------------------------------------------
//                              class c_xml_local
//              Stores local state associated with the XML parsing.
//          One thing it does is set up a local hierarchy for namespaces.
//------------------------------------------------------------------------------

class c_xml_local
{
private:
    const c_xml_local       *p_previous;
    c_xml_local             *p_active;      // Pointer to member that contains ns data
    c_repository<int>       namespaces;

public:
    LMX_PDECL explicit c_xml_local() : p_previous( LMXNULL ), p_active( LMXNULL ) {}
    LMX_PDECL c_xml_local( const c_xml_local &ar_previous ) :
            p_previous( &ar_previous ), p_active( ar_previous.p_active )
    {}
    LMX_PDECL ~c_xml_local() {}

    LMX_PDECL int find_namespace_id( const std::string &ar_namespace_prefix ) const;
    LMX_PDECL const char *find_namespace_prefix( int a_namespace_id ) const;
    LMX_PDECL const char *find_first_namespace_prefix( int a_namespace_id ) const;
    LMX_PDECL void add_namespace( const char *ar_namespace_prefix, int a_namespace_id );
    LMX_PDECL void clear() { namespaces.clear(); }
};

//------------------------------------------------------------------------------
//                              class c_attribute_data
//              Stores parsed attribute data.  Attributes data potentially
//           has to be parsed multiple time (to extract namespace information,
//          xsi:type, and finally the actual attribute data).  This store
//                            makes this more efficient.
//
// Note: attribute_data_note - The attribute_data structure becomes a
//          concatenation of sets of information describing the attribute data
//          for an element.  The first part of the set is the \0 terminated
//          attribute name.  The second part of the set is either ' or "
//          depending on the type of quotes around the attribute value.
//          (Storing this allows us to more accurately re-constitute the
//          attribute data when xs:any data is being retrieved.)  The last value
//          is the \0 terminated attribute value.  Entities are NOT expanded
//          when stored in this structure.  (This is again so we can
//          re-constitute xs:any data.)  If the attribute name parameter is
//          empty (i.e. just has \0), this marks the end of the data store.
//          An example of the stored data is therefore:
//
//                  name1\0"value1\0name2\0'value2\0\0
//
//          For convenience, each '\0' terminated bit is called a segment.
//------------------------------------------------------------------------------

class c_attribute_data
{
private:
    std::vector<char> data;         // See note attribute_data_note.
    size_t index;

public:
    c_attribute_data() : index(0) { data.reserve( 150 ); }
    ~c_attribute_data() {}

    inline void clear()                 { data.resize(0); index = 0; }      // Avoid deleting memory we have already allocated as we can reuse it
    inline bool empty() const           { return data.empty(); }
    inline bool is_eof() const          { return data.empty() || data[index] == '\0' && index + 1 >= data.size(); }
    inline void rewind()                { index = 0; }
    inline void put( char a_c )         { data.push_back( a_c ); }
    inline void append( const char *ap_begin, const char *ap_end )  // ap_end points to the char after the data
    { data.insert( data.end(), ap_begin, ap_end ); }
    inline void append( const std::string &ar_string )
    { data.insert( data.end(), ar_string.begin(), ar_string.end() ); }
    inline void mark_segment_end()      { put( '\0' ); }
    inline int peek() const             { if( data[index] == '\0' ) return c_read::k_eof; return (unsigned char)data[index]; }
    inline int get()                    { if( data[index] == '\0' ) return c_read::k_eof; return (unsigned char)data[index++]; }
    inline void next()                  { if( data[index] != '\0' ) ++index; }
    inline void unget( int a_c )
    { lmx_assert( index > 0 ); if( a_c != c_read::k_eof && index > 0 ) { --index; lmx_assert( data[index] == a_c ); } }
    inline void next_segment()          { while( data[index] != '\0' ) ++index; if( index + 1 < data.size() ) ++index; }
    inline size_t size() const          { return data.size(); }
    // N.B. The pointer returned by this function may become invalid if extra data is added to the store
    inline const char *get_str() const { return &data[index]; }     // String at current segment
    inline const char *get_str( size_t a_index ) const { return &data[a_index]; }   // String at specified segment
};

//------------------------------------------------------------------------------
//                               class c_xml
//                     This class understands XML things
//                     such as namespaces but not Schema.
//------------------------------------------------------------------------------

class c_xml
{
private:
    c_get_for_xml getter;   // c_get_for_xml getter MUST be declared before c_internal_dtd dtd
    std::string entity_tag;
    c_internal_dtd dtd;
    c_enumerated_namespaces enumerated_namespaces;
    bool dtd_read;
    bool element_read;
    mutable c_attribute_data attribute_data;    // To allow operation of get_xsi_type().
    c_attribute_data namespace_attribute_data;  // Stored so that we can re-constitute xs:any if required
    mutable std::string prefix_for_get_ns_id;

private:
    bool p_capture_attribute_value( c_attribute_data *ap_attribute_data, size_t *ap_initial_index=LMXNULL );
    bool p_capture_attributes( c_xml_local *ap_local_control = LMXNULL );
    bool p_read_cdata( std::string *ap_value );
    bool p_get_xml_text( std::string *ap_text );    // Read text part of mixed content
    bool p_handle_lt_qmark();       // Have read "<?"
    bool p_handle_lt_pling( std::string *ap_text, elxp_xml_event *ap_returnable_event );        // Have read "<!"

    // For xs:any handling
    enum e_tag_type { ETM_START, ETM_BODY, ETM_END, ETM_ERROR };

    void p_add_any_attribute_data( std::string *ap_value, c_attribute_data &ar_attribute_data );
    bool p_add_any_attributes( std::string *ap_value );
    e_tag_type p_get_any_tag_data( std::string *ap_value );
    bool p_get_any_element( const std::string &ar_name, std::string *ap_value );
    bool p_append_entity( const char *ap_entity_value, std::string *ap_value );
    bool p_select_entity( c_entity_name_reader &ar_tag_reader, std::string *ap_value );

public:
    LMX_PDECL c_xml( c_read &ar_reader, c_error &ar_error, bool a_is_reading_xsd = true );
    LMX_PDECL ~c_xml() {}

    LMX_PDECL void rewind();
    LMX_PDECL const char *get_source_id() const { return getter.get_source_id(); }
    LMX_PDECL int get_line_num() const { return getter.get_line_num(); }

    LMX_PDECL void add_general_entity( const std::string &ar_name, const std::string &ar_value );
    LMX_PDECL void add_parameter_entity( const std::string &ar_name, const std::string &ar_value );

    LMX_PDECL void add_namespace( const char *ap_namespace_uri, int a_value );
    LMX_PDECL int add_namespace( const char *ap_namespace_uri );
    LMX_PDECL int find_enumerated_namespace( const std::string &ar_namespace ) const;
    LMX_PDECL const char * find_namespace_from_id( int a_namespace_id ) const;

    LMX_PDECL static std::string & get_local_name( const std::string &ar_name, std::string *ap_local_name_out );
    LMX_PDECL static std::string & get_namespace_prefix( const std::string &ar_name, std::string *ap_namespace_prefix_out );
    LMX_PDECL int get_ns_id( const c_xml_local *ap_xml_local, const std::string &ar_name ) const;
    LMX_PDECL bool is_xsi_attribute( const c_xml_local *ap_xml_local, const std::string &ar_name ) const;

    LMX_PDECL elxp_xml_event get_element_name( c_xml_local *ap_local_control, std::string *ap_name );
                                // Reads an element name
                                // - also takes care of CDATA
    LMX_PDECL bool get_element_value( std::string *ap_value,
                                    elmx_whitespace a_ws_mode = EXWS_PRESERVE,
                                    bool a_expand_entities = true );
    LMX_PDECL bool is_element_end_tag( const std::string &ar_name );
    LMX_PDECL bool get_any_element( const std::string &ar_name, std::string *ap_value,
                                    elmx_whitespace a_ws_mode = EXWS_PRESERVE,
                                    bool a_expand_entities = false );
    LMX_PDECL bool get_nested_element_value( std::string *ap_value, std::string *ap_end_tag_name,
                                    elmx_whitespace a_ws_mode = EXWS_PRESERVE,
                                    bool a_expand_entities = false );

    LMX_PDECL elxp_xml_event get_attribute_name( std::string *ap_name, bool a_in_xml_decl = false );
    LMX_PDECL bool get_attribute_value( std::string *ap_value,
                                    elmx_whitespace a_ws_mode = EXWS_PRESERVE );
                                // Reads an attribute and does suitable expansion
                                // - Skip white space then look for " or '
                                // - Also does attribute normalisation
    LMX_PDECL bool skip_attribute_value();
    LMX_PDECL bool skip_start_tag( bool *ap_is_end_of_element = LMXNULL )
        { return getter.move_to_start_tag_end( ap_is_end_of_element ); }
    LMX_PDECL bool get_xsi_type( const c_xml_local *ap_xml_local,
                                int *ap_type_ns_id, std::string *ap_type_local_name );  // See get_xsi_type_is_non_const
    LMX_PDECL bool get_named_attribute(
                                const c_xml_local *ap_xml_local,
                                int a_ns_id,
                                const std::string &ar_local_name,
                                std::string *ap_value );    // V3.7

    LMX_PDECL bool skip_rest_of_element( const std::string &ar_name );
                                // Skips remainder of start tag, element body (if any)
                                // -and element end tag.
    LMX_PDECL bool skip_rest_of_empty_element( const std::string &ar_name );
    LMX_PDECL bool skip_element_body( const std::string &ar_name );
    LMX_PDECL static bool is_error_event( int a_event )
    {
        // Don't include EXE_EOM as a message that contains only a global simple type get this
        if( a_event == EXE_FATAL_ERROR || a_event == EXE_BADLY_FORMED )
            return true;
        return false;
    }
};

//------------------------------------------------------------------------------
//                           Tokeniser
//------------------------------------------------------------------------------

class c_xml_list_tokeniser
{
private:
    std::string o_string;
    std::string::const_iterator oi_position;

public:
    LMX_PDECL c_xml_list_tokeniser( const std::string &ar_string );
    LMX_PDECL ~c_xml_list_tokeniser() {}
    LMX_PDECL void set( const std::string &ar_string );
    LMX_PDECL bool get( std::string *ap_sub_string );
};

//------------------------------------------------------------------------------
//                  Reader and writer classes common components
//------------------------------------------------------------------------------

typedef int t_enumed_namespace;

struct s_event_map
{
    t_enumed_namespace          known_namespace;
    const char *                p_name;
    size_t                      size;
    t_xml_event                 event;
};

struct s_ns_map
{
    int                         id;
    const char *                p_ns_prefix;
    const char *                p_ns;
};

struct s_user_ns_prefix_map
{
    std::string ns_prefix;
    std::string ns_uri;
    s_user_ns_prefix_map( const std::string &ar_ns_prefix, const std::string &ar_ns_uri ) :
            ns_prefix( ar_ns_prefix ), ns_uri( ar_ns_uri )
    {}
};

typedef std::vector< s_user_ns_prefix_map > t_user_namespaces;

struct s_schema_location_item
{
    std::string ns_uri;
    std::string location;
    s_schema_location_item( const std::string &ar_ns_uri, const std::string &ar_location ) :
            ns_uri( ar_ns_uri ), location( ar_location )
    {}
};

typedef std::vector< s_schema_location_item > t_schema_locations;

//------------------------------------------------------------------------------
//                          struct s_debug_error
//  This structure provides a way to capture (un)marshalling error information
//  during code debugging.  It's prime purpose is to help debugging when the
//  convenience methods are being used.  If extended error information is
//  required in Release mode, then the Advanced forms of (un)marshalling should
//  be used.
//------------------------------------------------------------------------------

#if defined( _DEBUG ) && ! defined( LMX_DEBUG_CAPTURE_ERROR )
    #define LMX_DEBUG_CAPTURE_ERROR 1
#endif

#if defined( LMX_DEBUG_CAPTURE_ERROR ) && LMX_DEBUG_CAPTURE_ERROR == 1
struct s_debug_error
{
    elmx_error error;
    std::string item_name;
    const char *p_prog_file;
    int prog_line;
    int xml_line;

    s_debug_error();
    void clear();
    void set( elmx_error a_error, const std::string &ar_item_name,
                const char *ap_prog_file, int a_prog_line, int a_xml_line = -1 );
    std::ostream &to_stream( std::ostream &ar_os ) const;
};

LMX_PDECL extern s_debug_error debug_error;
#endif

//------------------------------------------------------------------------------
//                               Reader class(es)
//------------------------------------------------------------------------------

class c_xml_reader;

class c_xml_reader_local : public c_xml_local
{
private:
    c_xml_reader_local *p_prev;
    c_xml_reader *p_reader;

public:
    LMX_PDECL c_xml_reader_local( c_xml_reader *ap_reader );
    LMX_PDECL explicit c_xml_reader_local();
    LMX_PDECL void set_reader( c_xml_reader *ap_reader );
    LMX_PDECL ~c_xml_reader_local();
};

class c_cracked_name
{
private:
    std::string full_name;
    bool has_ns_prefix;
    int ns_id;
    std::string local_name;
    std::string ns_prefix;

public:
    LMX_PDECL c_cracked_name() {}
    LMX_PDECL c_cracked_name( c_xml_local *ap_reader_local, const std::string &ar_name ) { set( ap_reader_local, ar_name ); }
    LMX_PDECL void set( c_xml_local *ap_reader_local, const std::string &ar_name );
    LMX_PDECL const std::string &get_full_name() const { return full_name; }
    LMX_PDECL bool get_has_ns_prefix() const { return has_ns_prefix; }
    LMX_PDECL int get_ns_id() const { return ns_id; }
    LMX_PDECL const std::string &get_ns_prefix() const { return ns_prefix; }
    LMX_PDECL const std::string &get_local_name() const { return local_name; }
};

//------------------------------------------------------------------------------
//                   class c_string_stack and c_string_stack_node
//        These two classes work together to produce a stack of strings.
//       The intention is that as we move in and out of elements their names
//           can be stored efficiently on the stack by using memory that
//                     might have been previously allocated.
//------------------------------------------------------------------------------

class c_string_stack
{
private:
    std::vector< char > string_stack;

    // All the functionality is in c_string_stack_node
    friend class c_string_stack_node;

public:
    LMX_PDECL size_t get_stack_size() const { return string_stack.size(); } // For testing
};

class c_string_stack_node
{
private:
    c_string_stack &r_string_stack;
    size_t my_string_base_index;

public:
    LMX_PDECL c_string_stack_node( c_string_stack &ar_string_stack, const std::string &ar_string ) :
        r_string_stack( ar_string_stack ),
        my_string_base_index( ar_string_stack.string_stack.size() )
    {
        ar_string_stack.string_stack.reserve( r_string_stack.string_stack.size() + ar_string.size() + 1 );
        for( size_t l_i=0; l_i<ar_string.size(); ++l_i ) r_string_stack.string_stack.push_back( ar_string[l_i] );
        r_string_stack.string_stack.push_back( '\0' );
    }
    LMX_PDECL const char * get_str() const { return &r_string_stack.string_stack[my_string_base_index]; }
    LMX_PDECL bool operator == ( const std::string &ar_string ) const
    {
        return ar_string.size() <= r_string_stack.string_stack.size() - my_string_base_index - 1 &&
                r_string_stack.string_stack[my_string_base_index + ar_string.size()] == '\0' &&
                ar_string == &r_string_stack.string_stack[my_string_base_index];
    }
    LMX_PDECL bool operator != ( const std::string &ar_string ) const { return ! (*this == ar_string); }
    LMX_PDECL ~c_string_stack_node()
    {
        r_string_stack.string_stack.resize( my_string_base_index );
    }
};

//------------------------------------------------------------------------------
//                               class c_xml_reader
//                    The main reader class when unmarshaling
//------------------------------------------------------------------------------

class c_xml_reader : public c_xml
{
private:
    c_xml_reader_local *p_reader_local;
    c_xml_reader_local bottom_reader_local;
    c_error error;
    elmx_error final_error_code;
    std::string error_item_name;
    c_string_stack element_name_stack;
    c_cracked_name cracked_name;
    t_xml_event event;
    t_user_namespaces user_namespaces;

public: // These members allow reuse of the same string during unmarshalling.
        // This increases efficiency by removing the need for repeated allocation.
    std::string name;
    std::string value;
    std::string attr_name;

public:
    // User accessible functions
    LMX_PDECL c_xml_reader( c_read &ar_low_reader ) :
                    c_xml( ar_low_reader, error ),
                    p_reader_local( LMXNULL ),
                    bottom_reader_local(),
                    final_error_code( ELMX_OK )
    {
        bottom_reader_local.set_reader( this );
        p_reader_local = &bottom_reader_local;
    }
    LMX_PDECL virtual ~c_xml_reader() {}
    LMX_PDECL c_string_stack & get_element_name_stack() { return element_name_stack; }
    LMX_PDECL void add_namespace_mapping( const std::string &ar_ns_prefix, const std::string &ar_namespace );
    LMX_PDECL void add_default_namespace_mapping( const std::string &ar_namespace )
    {
        add_namespace_mapping( "", ar_namespace );
    }
    LMX_PDECL static std::string & get_local_name( const std::string &ar_name, std::string *ap_local_name_out )
        { return c_xml::get_local_name( ar_name, ap_local_name_out ); }
    LMX_PDECL static std::string & get_namespace_prefix( const std::string &ar_name, std::string *ap_namespace_prefix_out )
        { return c_xml::get_namespace_prefix( ar_name, ap_namespace_prefix_out ); }
    LMX_PDECL std::string & get_namespace( const std::string &ar_name, std::string *ap_namespace_out ) const;
    LMX_PDECL static const std::string k_namespace_not_found;
    LMX_PDECL elmx_error handle_error( elmx_error a_error, const std::string &ar_item_name, const char *a_prog_file, int a_prog_line );
    LMX_PDECL virtual elmx_error user_handle_error( elmx_error a_error ) { return a_error; }
    LMX_PDECL virtual elmx_error user_handle_error( elmx_error a_error, const std::string &ar_item_name,
                                            const char *a_prog_file, int a_prog_line )
    {
    #if defined( LMX_DEBUG_CAPTURE_ERROR ) && LMX_DEBUG_CAPTURE_ERROR == 1
        debug_error.set( a_error, ar_item_name, a_prog_file, a_prog_line, get_line_no() );
    #endif
    #if LMX_USE_EXCEPTIONS == 1
        if( a_error != ELMX_OK )    // Avoid any chance of an unreachable code warning
            throw c_lmx_reader_exception( a_error, get_line_no() );
    #endif
        return user_handle_error( a_error );    // For backwards compatibility
    }
    LMX_PDECL virtual std::string & get_error_message( std::string *ap_message ) const;
    LMX_PDECL int get_line_no() const { return c_xml::get_line_num(); }
    LMX_PDECL const std::string & get_error_item_name() const { return error_item_name; }

    LMX_PDECL void set_ns_map( const s_ns_map *ap_ns_map );

    // Parser functions
    LMX_PDECL elmx_error get_error() { return error.get_last_error(); }
    LMX_PDECL bool get_error_info( elmx_error_type *ap_type, elmx_error *ap_code, int *ap_line )
    { return error.get_error_info( ap_type, ap_code, ap_line ); }
    LMX_PDECL c_xml_reader_local *get_reader_local() const { return p_reader_local; }
    LMX_PDECL void set_reader_local( c_xml_reader_local * ap_reader_local ) { p_reader_local = ap_reader_local; }
    LMX_PDECL const std::string &get_full_name() const { return cracked_name.get_full_name(); }
    LMX_PDECL bool get_has_ns_prefix() const { return cracked_name.get_has_ns_prefix(); }
    LMX_PDECL int get_ns_id() const { return cracked_name.get_ns_id(); }
    LMX_PDECL int get_element_ns_id() const;
    LMX_PDECL const std::string &get_local_name() const { return cracked_name.get_local_name(); }
    LMX_PDECL void tokenise( const s_event_map ac_event_map[], elxp_attr_elem a_type );
    LMX_PDECL bool is_current_event_tokenisable() const { return event == EXE_TOKEN || event == EXE_UNKNOWN || event >= EXE_LAST_XML_EVENT; }
    LMX_PDECL bool is_xsi_attribute( const std::string &ar_name ) const { return c_xml::is_xsi_attribute( p_reader_local, ar_name ); }
    LMX_PDECL bool get_xsi_type( int *ap_type_ns_id, std::string *ap_type_local_name )
        { return c_xml::get_xsi_type( p_reader_local, ap_type_ns_id, ap_type_local_name ); }
    LMX_PDECL bool get_named_attribute( int a_ns_id, const std::string &ar_local_name, std::string *ap_value )  // V3.7
        { return c_xml::get_named_attribute( p_reader_local, a_ns_id, ar_local_name, ap_value ); }
    LMX_PDECL bool get_attribute_event( const s_event_map ac_event_map[], elmx_error *ap_error, std::string *ap_name );
    LMX_PDECL void get_element_event( const s_event_map ac_event_map[], elmx_error *ap_error, std::string *ap_name );
    LMX_PDECL t_xml_event get_current_event() const { return event; }
    LMX_PDECL void restore_event( t_xml_event a_event ) { event = a_event; }
    LMX_PDECL bool get_nested_element_value( std::string *ap_value, std::string *ap_end_tag_name,
                                    elmx_whitespace a_ws_mode = EXWS_PRESERVE,
                                    bool a_expand_entities = false );
    LMX_PDECL bool get_simple_type_value( const std::string &ar_name, elmx_whitespace a_ws_mode, std::string *ap_value, elmx_error *ap_error, bool *ap_is_empty_element );
    LMX_PDECL bool get_element_ur_value( const std::string &ar_name, std::string *ap_value, elmx_error *ap_error, bool *ap_is_empty_element );

    // V3.2.4
    LMX_PDECL bool is_valid_qname( const std::string &ar_value ) const;
    template< typename Tstring >
    inline void set_qname( ct_qname< Tstring > &ar_qname, const std::string &ar_value ) const
    {
        c_cracked_name l_cracked_name;
        l_cracked_name.set( p_reader_local, ar_value );
        Tstring l_tstring;
        convert( &l_tstring, l_cracked_name.get_local_name() );
        ar_qname.set( find_namespace_from_id( l_cracked_name.get_ns_id() ), l_tstring );
    }
    inline void set_qname( std::string &ar_qname, const std::string &ar_value ) const
            { ar_qname = ar_value; }
    inline void set_qname( std::wstring &ar_qname, const std::string &ar_value ) const
            { convert( &ar_qname, ar_value ); }

    // Deprecated - For backwards compatibility - these methods are no longer preferred
    bool is_tokenisable( t_xml_event a_event ) const { return a_event == EXE_TOKEN || a_event == EXE_UNKNOWN || a_event >= EXE_LAST_XML_EVENT; }
    t_xml_event tokenise( const s_event_map ac_event_map[], const std::string &ar_name, elxp_attr_elem a_type );
    t_xml_event get_attribute_event( const s_event_map ac_event_map[], std::string *ap_name );
    t_xml_event get_element_event( const s_event_map ac_event_map[], std::string *ap_name );
    bool get_end_tag( const s_event_map ac_event_map[], const std::string &ar_name,
                        t_xml_event *ap_event, elmx_error *ap_error );
};

class c_xml_reader_file : public c_xml_reader
{
private:
    c_read_file low_reader;

public:
    LMX_PDECL c_xml_reader_file( const char *ap_file ) :
                    c_xml_reader( low_reader ),
                    low_reader( ap_file )
    {
    }
#ifdef _MSC_VER
    LMX_PDECL c_xml_reader_file( const wchar_t *ap_file ) :
                    c_xml_reader( low_reader ),
                    low_reader( ap_file )
    {
    }
#endif
    LMX_PDECL c_xml_reader_file( LMXSTDCRT::FILE *ah_fin ) :
                    c_xml_reader( low_reader ),
                    low_reader( ah_fin )
    {
    }
    LMX_PDECL ~c_xml_reader_file() {}
    LMX_PDECL bool is_open() const
    {
        return low_reader.is_open();
    }
};

class c_xml_reader_memory : public c_xml_reader
{
private:
    c_read_memory low_reader;

public:
    LMX_PDECL c_xml_reader_memory( const char *ap_message, size_t a_message_length ) :
                    c_xml_reader( low_reader ),
                    low_reader( ap_message, a_message_length )
    {
    }
    LMX_PDECL ~c_xml_reader_memory() {}
    LMX_PDECL bool is_open() const
    {
        return low_reader.is_open();
    }
};

//------------------------------------------------------------------------------
//                               Writer class
//------------------------------------------------------------------------------

class c_xml_writer
{
protected:
    std::ostream &os;
    int o_indent;
    const char *p_tab;
    const char *p_nl;
    bool o_is_ns_attr_written;
    c_error error;
    elmx_error final_error_code;
    std::string error_item_name;
    const s_ns_map *p_ns_map_to_use;
    t_user_namespaces user_namespaces;
    t_schema_locations schema_locations;
    t_class_identity expected_class_identity;


public:
    typedef int t_writer_options;
    LMX_PDECL static const t_writer_options include_xml_decl;
    LMX_PDECL static const t_writer_options include_standalone;

public:
    // User accesible functions
    LMX_PDECL c_xml_writer( std::ostream &ar_os, const char *ap_tab = "\t", const char *ap_nl = "\n" );
    LMX_PDECL c_xml_writer( std::ostream &ar_os, t_writer_options a_options, const char *ap_tab = "\t", const char *ap_nl = "\n" );
    LMX_PDECL virtual ~c_xml_writer()
    {
    }
    LMX_PDECL void add_namespace( const std::string &ar_ns_prefix, const std::string &ar_namespace );
    LMX_PDECL void add_default_namespace( const std::string &ar_namespace )
    {
        add_namespace( "", ar_namespace );
    }
    LMX_PDECL void add_schema_location( const std::string &ar_ns_uri, const std::string &ar_location ); // v3.1
    LMX_PDECL void add_no_namespace_schema_location( const std::string &ar_location )   // v3.1
    {
        add_schema_location( "", ar_location );
    }
    LMX_PDECL elmx_error handle_error( elmx_error a_error, const std::string &ar_item_name, const char *a_prog_file, int a_prog_line );
    LMX_PDECL virtual elmx_error user_handle_error( elmx_error a_error ) { return a_error; }
    LMX_PDECL virtual elmx_error user_handle_error( elmx_error a_error, const std::string &ar_item_name,
                                            const char *a_prog_file, int a_prog_line )
    {
    #if defined( LMX_DEBUG_CAPTURE_ERROR ) && LMX_DEBUG_CAPTURE_ERROR == 1
        debug_error.set( a_error, ar_item_name, a_prog_file, a_prog_line );
    #endif
    #if LMX_USE_EXCEPTIONS == 1
        if( a_error != ELMX_OK )    // Avoid any chance of an unreachable code warning
            throw c_lmx_exception( a_error );
    #endif
        return user_handle_error( a_error );    // For backwards compatibility
    }
    LMX_PDECL virtual std::string & get_error_message( std::string *ap_message ) const;
    LMX_PDECL const std::string & get_error_item_name() const { return error_item_name; }

    // Parser functions
    LMX_PDECL bool get_error_info( elmx_error_type *ap_type, elmx_error *ap_code, int *ap_line )
    {
        return error.get_error_info( ap_type, ap_code, ap_line );
    }
    LMX_PDECL virtual const char * endl() // Users can override this method
    {
        return p_nl;
    }
    LMX_PDECL virtual void nest() // Users can override this method
    {
        os << p_tab;
    }
    LMX_PDECL std::ostream & operator()(bool a_do_indent=true)
    {
        if( a_do_indent )
            for( int i=0; i<o_indent; i++ )
                nest();
        return os;
    }
    LMX_PDECL void indent()
    {
        ++o_indent;
    }
    LMX_PDECL void undent()
    {
        if( o_indent > 0 )
            --o_indent;
    }
    LMX_PDECL void conditionally_select_ns_map( const s_ns_map * );
    LMX_PDECL void conditionally_write_ns_attrs( bool );
    LMX_PDECL void reenable_namespace_map_output() { o_is_ns_attr_written = false; }    // V3.7

    LMX_PDECL t_class_identity get_expected_class_identity() const
    {
        return expected_class_identity;
    }
    LMX_PDECL void set_expected_class_identity( t_class_identity a_expected_class_identity )
    {
        expected_class_identity = a_expected_class_identity;
    }
    LMX_PDECL void clear_expected_class_identity()
    {
        expected_class_identity = LMXNULL;
    }

    // V3.2.4
    LMX_PDECL const char *get_namespace_prefix( const std::string &ar_namespace ) const;
    // The make_qname_string functions create an un-escaped std::string.  This is returned
    // -in an instance of c_as_xml_string.  This trickers th output converters
    // -to perform any necessary XML entity escaping.
    template< typename Tstring >
    c_as_xml_string make_qname_string( const ct_qname< Tstring > &ar_qname ) const
    {
        const char *lp_ns_prefix = get_namespace_prefix( ar_qname.get_namespace() );
        lmx_assert( lp_ns_prefix );     // Assume user won't specifgy an unsupported namespace
        std::string l_qname( lp_ns_prefix );
        if( ! l_qname.empty() )
            l_qname += ":";
        std::string l_narrow_string;
        convert( &l_narrow_string, ar_qname.get_local_name() );
        l_qname += l_narrow_string;
        return c_as_xml_string( l_qname );
    }
    inline c_as_xml_string make_qname_string( const std::string &ar_qname ) const
            { return c_as_xml_string( ar_qname ); }
    inline c_as_xml_string make_qname_string( const std::wstring &ar_qname ) const
            { return c_as_xml_string( ar_qname ); }
};

//------------------------------------------------------------------------------
//                         Error code interpretation
//------------------------------------------------------------------------------

LMX_PDECL const char *get_error_description( elmx_error a_error_code );

//------------------------------------------------------------------------------
//                 Output converters for char and unsigned char
// Included here so that marshaling code can include them inline without
// lmxuser.h having to include <ostream>.
//------------------------------------------------------------------------------

template <>
inline std::ostream & operator << ( std::ostream & ar_os, const c_as_xml<unsigned char> &ar_v )
{
    ar_os << static_cast<unsigned int>( *(ar_v.p) );
    return ar_os;
}

template <>
inline std::ostream & operator << ( std::ostream & ar_os, const c_as_xml<signed char> &ar_v )
{
    ar_os << static_cast<int>( *(ar_v.p) );
    return ar_os;
}

//------------------------------------------------------------------------------
//                             Must Understand
//------------------------------------------------------------------------------

enum elmx_must_understand { EMU_OK, EMU_ERROR, EMU_MUST_UNDERSTAND };

typedef elmx_must_understand (*t_must_understand_test_function)( const std::string & ar_value );

LMX_PDECL elmx_must_understand bool_must_understand( const std::string &ar_value );

LMX_PDECL elmx_must_understand has_must_understand( c_xml_reader & ar_reader,
                            const std::string &ar_namespace,
                            const std::string &ar_name,
                            t_must_understand_test_function a_function = bool_must_understand );

//------------------------------------------------------------------------------
//                             get_instance_namespace
//------------------------------------------------------------------------------

LMX_PDECL bool get_instance_namespace( c_xml_reader & ar_reader, std::string *ap_namespace_out );

//------------------------------------------------------------------------------
//                Marshalling and Unmarshalling helper templates
//------------------------------------------------------------------------------

template< class T >
elmx_error marshal( const T &ar_item, const char ac_file_name[] )
{
    std::ofstream l_fout( ac_file_name );
    if( ! l_fout.is_open() )
        return ELMX_NO_FILE;
    c_xml_writer l_writer( l_fout, lmx::c_xml_writer::include_xml_decl );
    return ar_item.marshal( l_writer );
}

template< class T >
elmx_error marshal( const T &ar_item, std::string *ap_string )
{
    std::ostringstream l_sout;
    c_xml_writer l_writer( l_sout, lmx::c_xml_writer::include_xml_decl );
    elmx_error l_error = ar_item.marshal( l_writer );
    if( l_error == ELMX_OK )
        *ap_string = l_sout.str();
    return l_error;
}

template< class T >
elmx_error marshal( const T &ar_item, std::ostream &ar_sos )
{
    c_xml_writer l_writer( ar_sos, lmx::c_xml_writer::include_xml_decl );
    return ar_item.marshal( l_writer );
}

template< class T >
elmx_error unmarshal( T *ap_item, const char ac_file_name[] )
{
    c_xml_reader_file l_reader( ac_file_name );
    if( ! l_reader.is_open() )
        return ELMX_NO_FILE;
    return ap_item->unmarshal( l_reader );
}

template< class T >
elmx_error unmarshal( T *ap_item, const char *ap_memory, size_t a_memory_size )
{
    c_xml_reader_memory l_reader( ap_memory, a_memory_size );
    return ap_item->unmarshal( l_reader );
}

template< class T >
elmx_error unmarshal( T *ap_item, const std::string &ar_string )
{
    return unmarshal( ap_item, ar_string.data(), ar_string.size() );
}

template< class T >
elmx_error unmarshal_partial(
        c_xml_reader & ar_reader,
        T *ap_item,
        const s_ns_map ac_ns_map[],
        const std::string &ar_namespace,
        const std::string &ar_local_name )
{
    ar_reader.set_ns_map( ac_ns_map );
    std::string l_name;
    elmx_error l_error = lmx::ELMX_OK;
    c_xml_reader_local l_reader_local( &ar_reader );
    ar_reader.get_element_event( LMXNULL, &l_error, &l_name );
    if( l_error != lmx::ELMX_OK )
        return l_error;
    std::string l_element_namespace;
    std::string l_element_local_name;
    ar_reader.get_namespace( l_name, &l_element_namespace );
    ar_reader.get_local_name( l_name, &l_element_local_name );
    if( l_element_namespace != ar_namespace || l_element_local_name != ar_local_name )
        return lmx::ELMX_ELEMENT_NOT_FOUND;
    return ap_item->unmarshal( ar_reader, l_name );
}

template< class T >
bool /*is_found*/ p_unmarshal_find(
        c_xml_reader & ar_reader,
        elmx_error *ap_error,
        T *ap_item,
        const std::string &ar_sought_namespace,
        const std::string &ar_sought_local_name,
        const std::string &ar_element_name )
{
    std::string l_element_namespace;
    std::string l_element_local_name;
    ar_reader.get_namespace( ar_element_name, &l_element_namespace );
    ar_reader.get_local_name( ar_element_name, &l_element_local_name );
    if( l_element_local_name == ar_sought_local_name && l_element_namespace == ar_sought_namespace )
        {
        *ap_error = ap_item->unmarshal( ar_reader, ar_element_name );
        return true;
        }

    bool l_is_end_of_element;

    ar_reader.skip_start_tag( &l_is_end_of_element );
    if( l_is_end_of_element )
        return false;

    std::string l_name;

    for(;;)
        {
        c_xml_reader_local l_reader_local( &ar_reader );
        ar_reader.get_element_event( LMXNULL, ap_error, &l_name );

        if( *ap_error != ELMX_OK )
            return false;
        else if( ar_reader.get_current_event() == lmx::EXE_TOKEN )
            {
            if( p_unmarshal_find(
                    ar_reader,
                    ap_error,
                    ap_item,
                    ar_sought_namespace,
                    ar_sought_local_name,
                    l_name ) )
                return true;
            else if( *ap_error != ELMX_OK )
                return false;
            }
        else if( ar_reader.get_current_event() == EXE_ELEMENT_END )
            {
            if( l_name != ar_element_name )
                *ap_error = ELMX_BAD_ELEMENT_END;
            return false;
            }
        else if( ar_reader.get_current_event() == EXE_CDATA )
            {}  // No action
        else
            {
            *ap_error = ELMX_BADLY_FORMED;
            return false;
            }
        }
    assert(0);  // Shouldn't get to here
    return false;
}

template< class T >
bool /*is_found*/ unmarshal_find(
        c_xml_reader & ar_reader,
        elmx_error *ap_error,
        T *ap_item,
        const s_ns_map ac_ns_map[],
        const std::string &ar_sought_namespace,
        const std::string &ar_sought_local_name )
{
    ar_reader.set_ns_map( ac_ns_map );
    c_xml_reader_local l_reader_local( &ar_reader );
    std::string l_name;
    ar_reader.get_element_event( LMXNULL, ap_error, &l_name );
    if( ar_reader.get_current_event() != EXE_TOKEN )
        return false;
    *ap_error = ELMX_OK;
    return p_unmarshal_find(
                ar_reader,
                ap_error,
                ap_item,
                ar_sought_namespace,
                ar_sought_local_name,
                l_name );
}

//------------------------------------------------------------------------------
//                                 Miscellaneous
//------------------------------------------------------------------------------

LMX_PDECL void bswap( void *ap_lhs, void *ap_rhs, size_t size );

}; // End of namepsace lmx

#if defined( _MSC_VER ) && _MSC_VER >= 1400
#pragma warning (pop)  // Restore VC8 'safe' warnings
#endif

#endif  // LMXPARSE_H

//------------------------------------------------------------------------------
//                 Copyright (c) 2003-2007 Codalogic Ltd.
//------------------------------------------------------------------------------
