//------------------------------------------------------------------------------
//                 Copyright (c) 2003-2007 Codalogic Ltd.
//
// This file forms part of THE SUPPORTING SOFTWARE of Codalogic's LMX W3C
// Schema to C++ code generator product.  It may not be used independent of
// files legitimately generated and used by the LMX product.  See the LMX
// licence agreement for further restrictions on the use of this file.
//
// THIS FILE AND THE SOFTWARE CONTAINED HEREIN IS PROVIDED 'AS IS' AND COMES
// WITH NO WARRANTIES OF ANY KIND.
//
// CODALOGIC LTD SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY ANYONE
// OR ANYTHING DUE TO THE USE OF THIS FILE HOWEVER THEY MAY BE CAUSED.
//
// For more information on the LMX compiler, go to:
//
//     http://www.xml2cpp.com
//
// Also see:
//
//     http://www.codalogic.com
//
// Version: 3.6
//      - Added support for regular expressions in the pattern class.
//
// Version: 3.3
//      - Added QName helper functions.
//
// Version: 3.2.4
//      - Added the ct_qname template to handle QNames better.
//
// Version: 3.2.3
//      - Added ct_simple_list template for when handling multiple
//        element occurences with lists of simple types.
//      - Fixed handling of fractional seconds less than 10.0 in time,
//        datetime and duration types.
//
// Version: 3.2.1
//      - Fixed handling of fractional seconds less than 10.0 in time and
//        datetime types.
//
// Version: 3.1
//      - Added function to return the number of Unicode characters encoded
//        into a UTF-8 encoded string.
//
// Version: 3.0
//      - Used std:: versions of a number of C utility functions.
//
// Version: 3.0 (Beta)
//      - Moved xisxxxx() functions from this header file to lmxinternals.h
//      - Modified #include structure so that user can do #include "lmxtype.h"
//        without having to worry about including lmxuser.h instead.
//
// Version: 2.11
//      - No changes
//
// Version: 2.10
//      - c_big_int now returns correct results when compared against "-0"
//      - Now enforce the "C" locale for string to float conversions.
//      - Unicode handling extended beyond the Basic Multi-lingual Plane
//        to full Unicode range (0x-0x10ffff).
//
// Version: 2.9.2
//      - Added switches for MS VC 2005
//
// Version: 2.9
//      - No changes
//
// Version: 2.8
//      - Added string converters for std::string to std::string and
//        std::wstring to std::wstring.  These non-converters make it easier
//        to operate in a environment where _TCHAR type mechanisms are being
//        used.
//      - Added definition of LMX_TDECL if it is not already defined.  During
//        normal compilation this is defined in lmxuser.h, but the extra
//        definition helps MS intellisense work correctly.
//
// Version: 2.7
//      - Moved inittowstring from lmxtypes.cpp to lmxinternals.h to allow
//        control of wide vs. narrow string usage for customers without
//        supporting software source code.
//
// Version: 2.6
//
// Version: 2.5
//
// Acknowledgements:
//      Borland additions kindly provided by Jeff Hay of HYTEC, Inc.,
//      http://www.hytecinc.com.  Find 'BORLANDC' to see additions.
//------------------------------------------------------------------------------


#ifndef LMXUSER_H
// Due to dependencies between lmxuser.h and lmxtypes.h (mainly to do with 64-bit
// values and c_big_int) lmxtypes.h should be included via lmxuser.h.  This allows
// setting up of declspec and user preferences.  Including lmxuser.h outside the
// LMXTYPES_H inclusion guards allows lmxtypes.h to be included by lmxuser.h at
// the right place.  When this include returns LMXTYPES_H will be defined, and
// the remainder of this file will not be included again.
#include "lmxuser.h"
#endif

#ifndef LMXTYPES_H
#define LMXTYPES_H

#if defined( _MSC_VER ) && _MSC_VER > 1000
#pragma once
#pragma warning(disable:4786)
#endif

#if defined( _MSC_VER ) && _MSC_VER >= 1400
#pragma warning (push)
#pragma warning (disable:4996)  // Disable VC8 'safe' warnings
#endif

// To help MS intellisense
#if ! defined( LMX_TDECL )
#define LMX_TDECL
#endif

#include <string>
#include <cstdio>
#include <vector>
#include <deque>
#include <iosfwd>
#include <cfloat>
#include <cassert>

#if defined( LMX_VALIDATE_PATTERNS ) && LMX_VALIDATE_PATTERNS == 1
#include "lmxregex.h"
#endif

namespace lmx {

//------------------------------------------------------------------------------
//                      Basic validation prototypes
//------------------------------------------------------------------------------

LMX_TDECL std::wstring & vtostring( std::wstring & ar_dest, const std::string & ar_src );
LMX_TDECL std::string & vtostring( std::string & ar_dest, const std::string & ar_src );
LMX_TDECL size_t length( const std::string & ar_utf8_string );  // v3.1
LMX_TDECL bool is_min_length_ok( const std::string & ar_utf8_string, size_t a_length );  // v3.1
LMX_TDECL bool is_max_length_ok( const std::string & ar_utf8_string, size_t a_length );  // v3.1
LMX_TDECL bool is_valid_bool( const std::string & ar_s );
LMX_TDECL bool is_valid_float( const std::string & ar_s );
LMX_TDECL bool is_valid_integer( const std::string & value );
LMX_TDECL int total_digits( const std::string & ar_value );
LMX_TDECL int fraction_digits( const std::string & ar_value );

//------------------------------------------------------------------------------
//        Template for multiple occurrence of simple type lists - V3.2.3
//------------------------------------------------------------------------------

template< typename T >
class ct_simple_list
{
private:
    std::deque< T > store;  // Base on deque so that it can contain bool members

public:
    ct_simple_list() {}
    ct_simple_list( const ct_simple_list<T> &ar_rhs ) : store( ar_rhs.store ) {}
    virtual ~ct_simple_list() {}
    size_t size() const { return store.size(); }
    const T & get( size_t i ) const { return store[i]; }
    const T & back() const { return store.back(); }
    void set( size_t i, const T &a ) { return store[i] = a; }
    void append( const T &a ) { store.push_back( a ); }
    void clear() { store.erase(); }
    ct_simple_list<T> &operator =( const ct_simple_list<T> &ar_rhs )
        { ct_simple_list<T> l_temp( ar_rhs ); swap( l_temp ); return *this; }
    void swap( ct_simple_list<T> &ar_rhs ) { store.swap( ar_rhs.store ); }
};

//------------------------------------------------------------------------------
//                       Schema type specific classes
//------------------------------------------------------------------------------

class c_big_int
{
private:
    std::string value;

    void p_init( const char * ap_string );

public:
    LMX_TDECL c_big_int();
    LMX_TDECL c_big_int( const c_big_int & ar_big_int );
    LMX_TDECL c_big_int( const char * ap_string );
    LMX_TDECL c_big_int( const std::string &ar_string );
    LMX_TDECL c_big_int( long a_long );
    LMX_TDECL c_big_int & operator =( const c_big_int & ar_big_int );
    LMX_TDECL c_big_int & operator =( const char * ap_string );
    LMX_TDECL c_big_int & operator =( const std::string &ar_string );
    LMX_TDECL c_big_int & operator =( long a_long );
    LMX_TDECL ~c_big_int() {}
    LMX_TDECL void swap( c_big_int & ar_big_int ) { value.swap( ar_big_int.value ); }
    LMX_TDECL std::string get_as_string() const;
    // This routine is unable to convert to the maximum negative value
    LMX_TDECL long get( bool *ap_is_overflowed = LMXNULL ) const;
    LMX_TDECL double get_as_float() const;
    // Includes versions to prevent the need to create temporary big_ints and strings
    LMX_TDECL friend bool operator == ( const c_big_int &ar_lhs, const char *ap_rhs );
    LMX_TDECL friend bool operator != ( const c_big_int &ar_lhs, const char *ap_rhs );
    LMX_TDECL friend bool operator > ( const c_big_int &ar_lhs, const char *ap_rhs );
    LMX_TDECL friend bool operator >= ( const c_big_int &ar_lhs, const char *ap_rhs );
    LMX_TDECL friend bool operator < ( const c_big_int &ar_lhs, const char *ap_rhs );
    LMX_TDECL friend bool operator <= ( const c_big_int &ar_lhs, const char *ap_rhs );
    LMX_TDECL friend bool operator == ( const c_big_int &ar_lhs, const std::string &ar_value );
    LMX_TDECL friend bool operator != ( const c_big_int &ar_lhs, const std::string &ar_value );
    LMX_TDECL friend bool operator > ( const c_big_int &ar_lhs, const std::string &ar_value );
    LMX_TDECL friend bool operator >= ( const c_big_int &ar_lhs, const std::string &ar_value );
    LMX_TDECL friend bool operator < ( const c_big_int &ar_lhs, const std::string &ar_value );
    LMX_TDECL friend bool operator <= ( const c_big_int &ar_lhs, const std::string &ar_value );
    LMX_TDECL friend bool operator == ( const std::string &ar_value, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator != ( const std::string &ar_value, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator > ( const std::string &ar_value, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator >= ( const std::string &ar_value, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator < ( const std::string &ar_value, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator <= ( const std::string &ar_value, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator == ( const char *ap_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator != ( const char *ap_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator > ( const char *ap_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator >= ( const char *ap_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator < ( const char *ap_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator <= ( const char *ap_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator == ( const c_big_int &ar_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator != ( const c_big_int &ar_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator > ( const c_big_int &ar_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_big_int &ar_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator < ( const c_big_int &ar_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_big_int &ar_lhs, const c_big_int &ar_rhs );
    LMX_TDECL friend std::ostream & operator << ( std::ostream &ar_os, const c_big_int &ar_rhs );
    friend int test_c_big_int1();
    friend int test_c_big_int2();
};

class c_decimal
{
private:
    bool is_neg;
    c_big_int left;
    std::string right;

    void p_init( const char * );
    static int p_fraction_comp( const std::string & ar_lhs, const std::string & ar_rhs );

public:
    LMX_TDECL c_decimal();
    LMX_TDECL c_decimal( const c_decimal & );
    LMX_TDECL c_decimal( const std::string & );
    LMX_TDECL c_decimal( const char * );
    LMX_TDECL c_decimal( long a_value, int a_scale = 0 );
    LMX_TDECL c_decimal( double a_value, unsigned int a_precision = DBL_DIG );
    LMX_TDECL ~c_decimal() {}
    LMX_TDECL static bool is_valid( const std::string & );
    LMX_TDECL c_decimal & operator = ( const c_decimal & );
    LMX_TDECL c_decimal & operator = ( const char * );
    LMX_TDECL c_decimal & operator = ( const std::string & ar_s ) { return operator = ( ar_s.c_str() ); }
    LMX_TDECL c_decimal & operator = ( double a_value );
    LMX_TDECL void swap( c_decimal & );
    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL long get_integral( bool *ap_is_overflowed = LMXNULL ) const;
    LMX_TDECL void set_integral( long a_v );
    LMX_TDECL unsigned long get_fractional( int a_precision, bool *ap_is_overflowed = LMXNULL ) const;
    LMX_TDECL void set_fractional( unsigned long a_v, unsigned int a_precision );
    LMX_TDECL long get_scaled( int a_scale, bool *ap_is_overflowed = LMXNULL ) const;
    LMX_TDECL void set_scaled( long a_v, int a_scale = 0 );
    LMX_TDECL double get_as_float() const;
    LMX_TDECL void set( double a_value, unsigned int a_precision = DBL_DIG );
    LMX_TDECL friend bool operator == ( const c_decimal &, const c_decimal & );
    LMX_TDECL friend bool operator != ( const c_decimal &, const c_decimal & );
    LMX_TDECL friend bool operator > ( const c_decimal &, const c_decimal & );
    LMX_TDECL friend bool operator >= ( const c_decimal &, const c_decimal & );
    LMX_TDECL friend bool operator < ( const c_decimal &, const c_decimal & );
    LMX_TDECL friend bool operator <= ( const c_decimal &, const c_decimal & );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_decimal & );
    LMX_TDECL friend int test_c_decimal();
};

class c_duration
{
private:
    bool is_neg;
    unsigned int years;
    unsigned int months;
    unsigned int days;
    unsigned int hours;
    unsigned int mins;
    float secs;
    static const unsigned int day_mons_mins[];
    static const unsigned int day_mons_maxs[];

    bool p_is_indeterminate( const c_duration & ar_rhs ) const;
    bool p_is_greater_than( const c_duration & ar_rhs ) const;
    bool p_is_equal( const c_duration & ar_rhs ) const;

public:
    LMX_TDECL c_duration();
    // c_duration( const c_duration & );    // Default copy constructor & assignment operator is fine
    LMX_TDECL c_duration( const std::string & );
    LMX_TDECL c_duration( const char * );
    LMX_TDECL ~c_duration() {}
    LMX_TDECL friend bool operator == ( const c_duration & ar_lhs, const c_duration & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_duration & ar_lhs, const c_duration & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_duration & ar_lhs, const c_duration & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_duration & ar_lhs, const c_duration & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_duration & ar_lhs, const c_duration & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_duration & ar_lhs, const c_duration & ar_rhs );
    LMX_TDECL static bool is_valid( const char * );
    LMX_TDECL static bool is_valid( const std::string & ar_s ) { return is_valid( ar_s.c_str() ); }
    LMX_TDECL c_duration & operator = ( const char * );
    LMX_TDECL c_duration & operator = ( const std::string & ar_s ) { return operator = ( ar_s.c_str() ); }

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL bool get_is_neg() const { return is_neg; }
    LMX_TDECL unsigned int get_years() const { return years; }
    LMX_TDECL unsigned int get_months() const { return months; }
    LMX_TDECL unsigned int get_days() const { return days; }
    LMX_TDECL unsigned int get_hours() const { return hours; }
    LMX_TDECL unsigned int get_mins() const { return mins; }
    LMX_TDECL float get_secs() const { return secs; }

    LMX_TDECL void set_is_neg( bool a_is_neg ) { is_neg = a_is_neg; }
    LMX_TDECL void set_years( unsigned int a_years ) { years = a_years; }
    LMX_TDECL void set_months( unsigned int a_months ) { months = a_months; }
    LMX_TDECL void set_days( unsigned int a_days ) { days = a_days; }
    LMX_TDECL void set_hours( unsigned int a_hours ) { hours = a_hours; }
    LMX_TDECL void set_mins( unsigned int a_mins ) { mins = a_mins; }
    LMX_TDECL void set_secs( float a_secs ) { lmx_assert( a_secs >= 0.0 ); secs = a_secs; }

    LMX_TDECL void normalize();

    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_duration & );
    LMX_TDECL friend int test_c_duration();
};

class c_datetime_base
{
protected:
    int year;      // Year can be negative
    unsigned int month;
    unsigned int day;
    unsigned int hours;
    unsigned int mins;
    float secs;
    int off_mins;   // Combination of offset hours and offset minutes
    bool is_utc;

public:
    LMX_TDECL c_datetime_base();

protected:
    bool is_equal( const c_datetime_base &ar_rhs ) const;
    bool is_greater_than( const c_datetime_base &ar_rhs ) const;
    void set_tz( const char *ap_s );
    static bool is_valid_tz( const char * & arp_s );
    std::ostream & output_tz( std::ostream &ar_os ) const;
    void set_offset( int a_off_hours, unsigned int a_off_mins )
        {
        if( a_off_hours >= 0 )
            off_mins = a_off_hours * 60 + a_off_mins;
        else
            off_mins = a_off_hours * 60 - a_off_mins;
        if( off_mins != 0 ) is_utc = false;
        }

    ~c_datetime_base() {}
};

class c_datetime : protected c_datetime_base
{
public:
    LMX_TDECL c_datetime();
    // c_datetime( const c_datetime & );                    // Default OK
    LMX_TDECL c_datetime( const char * );
    LMX_TDECL c_datetime( const std::string & );
    LMX_TDECL static bool is_valid( const std::string & );
    // c_datetime & operator = ( const c_datetime & );  // Default OK
    LMX_TDECL c_datetime & operator = ( const char * );
    LMX_TDECL c_datetime & operator = ( const std::string & );
    LMX_TDECL ~c_datetime() {}

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL int get_year() const { return year; }
    LMX_TDECL unsigned int get_month() const { return month; }
    LMX_TDECL unsigned int get_day() const { return day; }
    LMX_TDECL unsigned int get_hours() const { return hours; }
    LMX_TDECL unsigned int get_mins() const { return mins; }
    LMX_TDECL float get_secs() const { return secs; }
    LMX_TDECL int get_total_offset_mins() const { return off_mins; }  // Hours * 60 + Mins
    LMX_TDECL bool get_is_utc() const { return is_utc; }

    LMX_TDECL void set_year( int a_year ) { year = a_year; }
    LMX_TDECL void set_month( unsigned int a_month ) { month = a_month; }
    LMX_TDECL void set_day( unsigned int a_day ) { day = a_day; }
    LMX_TDECL void set_hours( unsigned int a_hours ) { hours = a_hours; }
    LMX_TDECL void set_mins( unsigned int a_mins ) { mins = a_mins; }
    LMX_TDECL void set_secs( float a_secs ) { lmx_assert( a_secs >= 0.0 ); secs = a_secs; }
    LMX_TDECL void set_total_offset_mins( int a_off_mins ) { off_mins = a_off_mins; } // Hours * 60 + Mins
    LMX_TDECL void set_offset_mins( int a_off_hours, unsigned int a_off_mins )    // Deprecated
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_offset( int a_off_hours, unsigned int a_off_mins )
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_is_utc( bool a_is_utc ) { is_utc = a_is_utc; }

    LMX_TDECL friend bool operator == ( const c_datetime & ar_lhs, const c_datetime & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_datetime & ar_lhs, const c_datetime & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_datetime & ar_lhs, const c_datetime & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_datetime & ar_lhs, const c_datetime & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_datetime & ar_lhs, const c_datetime & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_datetime & ar_lhs, const c_datetime & ar_rhs );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_datetime & );
};

class c_time : protected c_datetime_base
{
public:
    LMX_TDECL c_time();
    // c_time( const c_time & );                    // Default OK
    LMX_TDECL c_time( const char * );
    LMX_TDECL c_time( const std::string & );
    LMX_TDECL ~c_time() {}
    LMX_TDECL static bool is_valid( const std::string & );
    // c_time & operator = ( const c_time & );  // Default OK
    LMX_TDECL c_time & operator = ( const char * );
    LMX_TDECL c_time & operator = ( const std::string & );

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL unsigned int get_hours() const { return hours; }
    LMX_TDECL unsigned int get_mins() const { return mins; }
    LMX_TDECL float get_secs() const { return secs; }
    LMX_TDECL int get_total_offset_mins() const { return off_mins; }  // Hours * 60 + Mins
    LMX_TDECL bool get_is_utc() const { return is_utc; }

    LMX_TDECL void set_hours( unsigned int a_hours ) { hours = a_hours; }
    LMX_TDECL void set_mins( unsigned int a_mins ) { mins = a_mins; }
    LMX_TDECL void set_secs( float a_secs ) { lmx_assert( a_secs >= 0.0 ); secs = a_secs; }
    LMX_TDECL void set_total_offset_mins( int a_off_mins ) { off_mins = a_off_mins; }
    LMX_TDECL void set_offset_mins( int a_off_hours, unsigned int a_off_mins )    // Deprecated
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_offset( int a_off_hours, unsigned int a_off_mins )
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_is_utc( bool a_is_utc ) { is_utc = a_is_utc; }

    LMX_TDECL friend bool operator == ( const c_time & ar_lhs, const c_time & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_time & ar_lhs, const c_time & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_time & ar_lhs, const c_time & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_time & ar_lhs, const c_time & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_time & ar_lhs, const c_time & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_time & ar_lhs, const c_time & ar_rhs );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_time & );
};

class c_date : protected c_datetime_base
{
public:
    LMX_TDECL c_date();
    // c_date( const c_date & );                    // Default OK
    LMX_TDECL c_date( const char * );
    LMX_TDECL c_date( const std::string & );
    LMX_TDECL static bool is_valid( const std::string & );
    // c_date & operator = ( const c_date & );  // Default OK
    LMX_TDECL ~c_date() {}
    LMX_TDECL c_date & operator = ( const char * );
    LMX_TDECL c_date & operator = ( const std::string & );

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL int get_year() const { return year; }
    LMX_TDECL unsigned int get_month() const { return month; }
    LMX_TDECL unsigned int get_day() const { return day; }
    LMX_TDECL int get_total_offset_mins() const { return off_mins; }  // Hours * 60 + Mins
    LMX_TDECL bool get_is_utc() const { return is_utc; }

    LMX_TDECL void set_year( int a_year ) { year = a_year; }
    LMX_TDECL void set_month( unsigned int a_month ) { month = a_month; }
    LMX_TDECL void set_day( unsigned int a_day ) { day = a_day; }
    LMX_TDECL void set_total_offset_mins( int a_off_mins ) { off_mins = a_off_mins; }
    LMX_TDECL void set_offset_mins( int a_off_hours, unsigned int a_off_mins )    // Deprecated
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_offset( int a_off_hours, unsigned int a_off_mins )
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_is_utc( bool a_is_utc ) { is_utc = a_is_utc; }

    LMX_TDECL friend bool operator == ( const c_date & ar_lhs, const c_date & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_date & ar_lhs, const c_date & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_date & ar_lhs, const c_date & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_date & ar_lhs, const c_date & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_date & ar_lhs, const c_date & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_date & ar_lhs, const c_date & ar_rhs );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_date & );
};

class c_gyearmonth : protected c_datetime_base
{
public:
    LMX_TDECL c_gyearmonth();
    // c_gyearmonth( const c_gyearmonth & );                // Default OK
    LMX_TDECL c_gyearmonth( const char * );
    LMX_TDECL c_gyearmonth( const std::string & );
    LMX_TDECL ~c_gyearmonth() {}
    LMX_TDECL static bool is_valid( const std::string & );
    // c_gyearmonth & operator = ( const c_gyearmonth & );// Default OK
    LMX_TDECL c_gyearmonth & operator = ( const char * );
    LMX_TDECL c_gyearmonth & operator = ( const std::string & );

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL int get_year() const { return year; }
    LMX_TDECL unsigned int get_month() const { return month; }
    LMX_TDECL int get_total_offset_mins() const { return off_mins; }          // Hours * 60 + Mins
    LMX_TDECL bool get_is_utc() const { return is_utc; }

    LMX_TDECL void set_year( int a_year ) { year = a_year; }
    LMX_TDECL void set_month( unsigned int a_month ) { month = a_month; }
    LMX_TDECL void set_total_offset_mins( int a_off_mins ) { off_mins = a_off_mins; }
    LMX_TDECL void set_offset_mins( int a_off_hours, unsigned int a_off_mins )    // Deprecated
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_offset( int a_off_hours, unsigned int a_off_mins )
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_is_utc( bool a_is_utc ) { is_utc = a_is_utc; }

    LMX_TDECL friend bool operator == ( const c_gyearmonth & ar_lhs, const c_gyearmonth & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_gyearmonth & ar_lhs, const c_gyearmonth & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_gyearmonth & ar_lhs, const c_gyearmonth & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_gyearmonth & ar_lhs, const c_gyearmonth & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_gyearmonth & ar_lhs, const c_gyearmonth & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_gyearmonth & ar_lhs, const c_gyearmonth & ar_rhs );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_gyearmonth & );
};

class c_gyear : protected c_datetime_base
{
public:
    LMX_TDECL c_gyear();
    // c_gyear( const c_gyear & );              // Default OK
    LMX_TDECL c_gyear( const char * );
    LMX_TDECL c_gyear( const std::string & );
    LMX_TDECL ~c_gyear() {}
    LMX_TDECL static bool is_valid( const std::string & );
    // c_gyear & operator = ( const cmlx_gyear & ); // Default OK
    LMX_TDECL c_gyear & operator = ( const char * );
    LMX_TDECL c_gyear & operator = ( const std::string & );

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL int get_year() const { return year; }
    LMX_TDECL int get_total_offset_mins() const { return off_mins; }  // Hours * 60 + Mins
    LMX_TDECL bool get_is_utc() const { return is_utc; }

    LMX_TDECL void set_year( int a_year ) { year = a_year; }
    LMX_TDECL void set_total_offset_mins( int a_off_mins ) { off_mins = a_off_mins; }
    LMX_TDECL void set_offset_mins( int a_off_hours, unsigned int a_off_mins )    // Deprecated
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_offset( int a_off_hours, unsigned int a_off_mins )
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_is_utc( bool a_is_utc ) { is_utc = a_is_utc; }

    LMX_TDECL friend bool operator == ( const c_gyear & ar_lhs, const c_gyear & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_gyear & ar_lhs, const c_gyear & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_gyear & ar_lhs, const c_gyear & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_gyear & ar_lhs, const c_gyear & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_gyear & ar_lhs, const c_gyear & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_gyear & ar_lhs, const c_gyear & ar_rhs );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_gyear& );
};

class c_gmonthday : protected c_datetime_base
{
public:
    LMX_TDECL c_gmonthday();
    // c_gmonthday( const c_gmonthday & );              // Default OK
    LMX_TDECL c_gmonthday( const char * );
    LMX_TDECL c_gmonthday( const std::string & );
    LMX_TDECL ~c_gmonthday() {}
    LMX_TDECL static bool is_valid( const std::string & );
    // c_gmonthday & operator = ( const c_gmonthday & );    // Default OK
    LMX_TDECL c_gmonthday & operator = ( const char * );
    LMX_TDECL c_gmonthday & operator = ( const std::string & );

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL unsigned int get_month() const { return month; }
    LMX_TDECL unsigned int get_day() const { return day; }
    LMX_TDECL int get_total_offset_mins() const { return off_mins; }          // Hours * 60 + Mins
    LMX_TDECL bool get_is_utc() const { return is_utc; }

    LMX_TDECL void set_month( unsigned int a_month ) { month = a_month; }
    LMX_TDECL void set_day( unsigned int a_day ) { day = a_day; }
    LMX_TDECL void set_total_offset_mins( int a_off_mins ) { off_mins = a_off_mins; }
    LMX_TDECL void set_offset_mins( int a_off_hours, unsigned int a_off_mins )    // Deprecated
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_offset( int a_off_hours, unsigned int a_off_mins )
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_is_utc( bool a_is_utc ) { is_utc = a_is_utc; }

    LMX_TDECL friend bool operator == ( const c_gmonthday & ar_lhs, const c_gmonthday & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_gmonthday & ar_lhs, const c_gmonthday & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_gmonthday & ar_lhs, const c_gmonthday & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_gmonthday & ar_lhs, const c_gmonthday & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_gmonthday & ar_lhs, const c_gmonthday & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_gmonthday & ar_lhs, const c_gmonthday & ar_rhs );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_gmonthday & );
};

class c_gday : protected c_datetime_base
{
public:
    LMX_TDECL c_gday();
    // c_gday( const c_gday & );                    // Default OK
    LMX_TDECL c_gday( const char * );
    LMX_TDECL c_gday( const std::string & );
    LMX_TDECL ~c_gday() {}
    LMX_TDECL static bool is_valid( const std::string & );
    // c_gday & operator = ( const c_gday & );  // Default OK
    LMX_TDECL c_gday & operator = ( const char * );
    LMX_TDECL c_gday & operator = ( const std::string & );

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL unsigned int get_day() const { return day; }
    LMX_TDECL int get_total_offset_mins() const { return off_mins; }  // Hours * 60 + Mins
    LMX_TDECL bool get_is_utc() const { return is_utc; }

    LMX_TDECL void set_day( unsigned int a_day ) { day = a_day; }
    LMX_TDECL void set_total_offset_mins( int a_off_mins ) { off_mins = a_off_mins; }
    LMX_TDECL void set_offset_mins( int a_off_hours, unsigned int a_off_mins )    // Deprecated
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_offset( int a_off_hours, unsigned int a_off_mins )
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_is_utc( bool a_is_utc ) { is_utc = a_is_utc; }

    LMX_TDECL friend bool operator == ( const c_gday & ar_lhs, const c_gday & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_gday & ar_lhs, const c_gday & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_gday & ar_lhs, const c_gday & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_gday & ar_lhs, const c_gday & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_gday & ar_lhs, const c_gday & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_gday & ar_lhs, const c_gday & ar_rhs );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_gday & );
};

class c_gmonth : protected c_datetime_base
{
public:
    LMX_TDECL c_gmonth();
    // c_gmonth( const c_gmonth & );                    // Default OK
    LMX_TDECL c_gmonth( const char * );
    LMX_TDECL c_gmonth( const std::string & );
    LMX_TDECL ~c_gmonth() {}
    LMX_TDECL static bool is_valid( const std::string & );
    // c_gmonth & operator = ( const c_gmonth & );  // Default OK
    LMX_TDECL c_gmonth & operator = ( const char * );
    LMX_TDECL c_gmonth & operator = ( const std::string & );

    LMX_TDECL std::string get_as_string() const;
    LMX_TDECL unsigned int get_month() const { return month; }
    LMX_TDECL int get_total_offset_mins() const { return off_mins; }      // Hours * 60 + Mins
    LMX_TDECL bool get_is_utc() const { return is_utc; }

    LMX_TDECL void set_month( unsigned int a_month ) { month = a_month; }
    LMX_TDECL void set_total_offset_mins( int a_off_mins ) { off_mins = a_off_mins; }
    LMX_TDECL void set_offset_mins( int a_off_hours, unsigned int a_off_mins )    // Deprecated
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_offset( int a_off_hours, unsigned int a_off_mins )
        { c_datetime_base::set_offset( a_off_hours, a_off_mins ); }
    LMX_TDECL void set_is_utc( bool a_is_utc ) { is_utc = a_is_utc; }

    LMX_TDECL friend bool operator == ( const c_gmonth & ar_lhs, const c_gmonth & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_gmonth & ar_lhs, const c_gmonth & ar_rhs );
    LMX_TDECL friend bool operator > ( const c_gmonth & ar_lhs, const c_gmonth & ar_rhs );
    LMX_TDECL friend bool operator >= ( const c_gmonth & ar_lhs, const c_gmonth & ar_rhs );
    LMX_TDECL friend bool operator < ( const c_gmonth & ar_lhs, const c_gmonth & ar_rhs );
    LMX_TDECL friend bool operator <= ( const c_gmonth & ar_lhs, const c_gmonth & ar_rhs );
    LMX_TDECL friend std::ostream & operator << (std::ostream &, const c_gmonth & );
};

class c_binary
{
private:
    std::vector<unsigned char> value;

    static unsigned char p_base64_char_to_6_bits( int a_c );
    void p_base64_convert_24bits_to_bytes( tlmx_uns32 a_accumulated, size_t a_number_accumulated );
    static char p_6_bits_to_base64_char( int a_c );
    static void p_convert_24bits_to_4x6bits( std::ostream &ar_os, tlmx_uns32 a_accumulated, size_t a_number_accumulated );

public:
    LMX_TDECL c_binary() {}
    LMX_TDECL c_binary( const c_binary & ar_v );
    LMX_TDECL c_binary( const unsigned char *ap_v, size_t a_len );
    LMX_TDECL ~c_binary() {}
    LMX_TDECL c_binary & operator = ( const c_binary & );
    LMX_TDECL void swap( c_binary & ar_v );
    LMX_TDECL void set( const unsigned char *ap_v, size_t a_len );
    LMX_TDECL static bool is_valid_hex( const std::string & ar_v );
    LMX_TDECL static bool is_valid_base64( const std::string & ar_v );
    LMX_TDECL void from_hex( const std::string & ar_src );
    LMX_TDECL void from_base64( const std::string & ar_src );
    LMX_TDECL unsigned char operator[] ( size_t a_i ) const;
    LMX_TDECL unsigned char & operator[] ( size_t a_i );
    #if defined( _MSC_VER ) ||  defined( __GNUC__ ) && __GNUC__ >= 3
    // pre-GCC v3 does not have vector::at (at() throws an exception if out of bounds)
    LMX_TDECL unsigned char at( size_t a_i ) const;
    LMX_TDECL unsigned char & at( size_t a_i );
    #endif
    LMX_TDECL void append( unsigned char a_c );
    LMX_TDECL void operator += ( unsigned char a_c );
    LMX_TDECL size_t size() const;
    LMX_TDECL unsigned char *get( unsigned char *ap_buffer, size_t a_buffer_size ) const;
    LMX_TDECL void out_hex( std::ostream & ar_os ) const;
    LMX_TDECL void out_base64( std::ostream & ar_os ) const;
    LMX_TDECL bool equal( const unsigned char ac_rhs[], size_t a_rhs_len ) const;
    LMX_TDECL friend bool operator == ( const c_binary & ar_lhs, const c_binary & ar_rhs );
    LMX_TDECL friend bool operator != ( const c_binary & ar_lhs, const c_binary & ar_rhs );
    LMX_TDECL friend int test_c_binary();
};

template< typename Tstring >    // V3.2.4
class ct_qname
{
private:
    Tstring local_name;
    std::string o_namespace;

public:
    ct_qname() {}
    ct_qname( const ct_qname< Tstring > &ar_rhs )
            { local_name = ar_rhs.local_name; o_namespace = ar_rhs.o_namespace; }
    ct_qname< Tstring > & operator = ( const ct_qname< Tstring > & ar_rhs )
            { ct_qname< Tstring > t( ar_rhs ); swap( t ); return *this; }
    void swap( ct_qname< Tstring > & ar_rhs )
            { local_name.swap( ar_rhs.local_name ); o_namespace.swap( ar_rhs.o_namespace ); }
    void set( const std::string &ar_namespace, const Tstring &ar_local_name )
            { o_namespace = ar_namespace; local_name = ar_local_name; }
    void set_local_name( const Tstring &ar_local_name )
            { local_name = ar_local_name; }
    const Tstring &get_local_name() const
            { return local_name; }
    void set_namespace( const std::string &ar_namespace )
            { o_namespace = ar_namespace; }
    void set_namespace( const char *ap_namespace )
            { o_namespace = ap_namespace; }
    const std::string &get_namespace() const
            { return o_namespace; }
    bool operator == ( const ct_qname< Tstring > &ar_rhs ) const
            { return local_name == ar_rhs.local_name && o_namespace == ar_rhs.o_namespace; }
    bool operator != ( const ct_qname< Tstring > &ar_rhs ) const
            { return ! (*this == ar_rhs); }
    ~ct_qname() {}
};

LMX_TDECL void make_qname( ct_qname< std::string > *ap_value, const std::string &ar_qname, const std::string &ar_namespace, const std::string &ar_local_name );
LMX_TDECL void make_qname( ct_qname< std::wstring > *ap_value, const std::string &ar_qname, const std::string &ar_namespace, const std::string &ar_local_name );
LMX_TDECL void make_qname( std::string *ap_value, const std::string &ar_qname, const std::string &ar_namespace, const std::string &ar_local_name );
LMX_TDECL void make_qname( std::wstring *ap_value, const std::string &ar_qname, const std::string &ar_namespace, const std::string &ar_local_name );

//------------------------------------------------------------------------------
//                            Conversion functions
//------------------------------------------------------------------------------

LMX_TDECL std::string &convert( std::string *ap_out, const std::wstring &ar_in );
LMX_TDECL std::wstring &convert( std::wstring *ap_out, const std::string &ar_in );
inline std::string &convert( std::string *ap_out, const std::string &ar_in ) // V2.8
{
    *ap_out = ar_in;
    return *ap_out;
}
inline std::wstring &convert( std::wstring *ap_out, const std::wstring &ar_in ) // V2.8
{
    *ap_out = ar_in;
    return *ap_out;
}

//------------------------------------------------------------------------------
//                     Pattern Facet Validation Classes
// These are the default class that does pattern facet validation.
// c_pattern_null is included for backwards compatibility with old versions of
// LMX that did not do pattern validation.  Set #define LMX_VALIDATE_PATTERNS
// to 0 to use c_pattern_null.
//------------------------------------------------------------------------------

#if defined( LMX_VALIDATE_PATTERNS ) && LMX_VALIDATE_PATTERNS == 1
class c_pattern
{
private:
    c_regex regex;

public:
    LMX_TDECL c_pattern( const char *ap_pattern ) : regex( ap_pattern ) {}
    LMX_TDECL bool is_match( const std::string &ar_v ) const
    { return regex.is_match( ar_v ); }
    LMX_TDECL bool is_match( const std::wstring &ar_v ) const
    { std::string l_narrow; return regex.is_match( convert( &l_narrow, ar_v ) ); }
};
#endif

class c_pattern_null
{
public:
    c_pattern_null( const char *ap_pattern ) {}
    bool is_match( const std::string &ar_v ) const { return true; }
    bool is_match( const std::wstring &ar_v ) const { return true; }
};


} // End of namespace lmx

#if defined( _MSC_VER ) && _MSC_VER >= 1400
#pragma warning (pop)  // Restore VC8 'safe' warnings
#endif

#endif // LMXTYPES_H


//------------------------------------------------------------------------------
//                 Copyright (c) 2003-2007 Codalogic Ltd.
//------------------------------------------------------------------------------
