//------------------------------------------------------------------------------
//                 Copyright (c) 2003-2007 Codalogic Ltd.
//
// This file forms part of THE SUPPORTING SOFTWARE of Codalogic's LMX W3C
// Schema to C++ code generator product.  It may not be used independent of
// files legitimately generated and used by the LMX product.  See the LMX
// license agreement for further restrictions on the use of this file.
//
// Users may edit the contents of this file to suit their platform needs
// subject to the aforementioned restrictions remaining in effect on the
// derived work.
//
// THIS FILE AND THE SOFTWARE CONTAINED HEREIN IS PROVIDED 'AS IS' AND COMES
// WITH NO WARRANTIES OF ANY KIND.
//
// CODALOGIC LTD SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY ANYONE
// OR ANYTHING DUE TO THE USE OF THIS FILE HOWEVER THEY MAY BE CAUSED.
//
// For more information on the LMX compiler, go to:
//
//     http://www.xml2cpp.com
//
// Also see:
//
//     http://www.codalogic.com
//
// Version: 3.7
//      - Added manipulator to enable output of SOAP like boolean values that
//        are either 0 or 1.
//
// Version: 3.6
//      - Made the choice between using narrow and wide strings on the user
//        side a #defineable option.
//      - Defined tlmx_unicode_string as a replacement for tlmx_wide_string
//        as it does not make sense to be able to define a wide string as
//        narrow!
//      - Added support for specifying how pattern facets are handled.  See
//        USER: VALIDATE_PATTERNS for more information.
//
// Version: 3.3
//      - Added LL and ULL qualifiers to GCC versions of MKINT64 and MKUNS64.
//      - Added user configuration of LMX_XML_STRICT and set the default to 1.
//
// Version 3.2.1
//      - Changed position in file of lmx_assert() macro.
//
// Version 3.0
//      - Modified logic for determining Windows declspecs.
//      - Added switches to allow operation under Windows CE.
//      - Used std:: versions of a number of C utility functions.
//
// Version 3.0 (Beta)
//      - Changed uses of NULL to LMXNULL using suitable #define.
//
// Version: 2.11.1
//      - Allowed for setting some #define options on the C++ compiler
//        command-line.
//
// Version: 2.11
//      - No changes
//
// Version: 2.10
//      - Now enforce the "C" locale for string to float conversions.
//      - Unicode handling extended beyond the Basic Multi-lingual Plane
//        to full Unicode range (0x-0x10ffff).
//
// Version: 2.9.2
//      - Added switches for MS VC 2005
//
// Version: 2.9
//      - Modified the way tlmx_int64 can conditionally be represented as
//        c_big_int.  See LMX_REPRESENT_INT64_USING_C_BIG_INT for changes.
//      - Moved float_zero() function into lmxparse.cpp to reduce chance of
//        divide by zero error.
//
// Version: 2.8
//
// Version: 2.7         - search for (V2.7) to find changes in file.
//      - Moved position where lmxinternals.h is included.
//      - Remove definition of inittowstring.  This is now an inline
//        function in lmxinternals.h.
//      - Included lmx_assert() macro that allows tailoring of handling
//        of assertion conditions.
//
// Version: 2.6
//
// Version: 2.5
//
// Acknowledgements:
//      Borland additions kindly provided by Jeff Hay of HYTEC, Inc.,
//      http://www.hytecinc.com.  Find 'BORLANDC' to see additions.
//------------------------------------------------------------------------------


// Portability: IEEE754 - In order to set the IEEE 754 values NaN, INF and -INF
//              the code uses operations derived from IEEE 754.  As part of
//              this, it also assumes that exceptions are not signalled when a
//              divide by 0.0 is done.  If your platform does not satisfy
//              these requirements you may have to change the code marked as
//              IEEE754.

#ifndef LMXUSER_H
#define LMXUSER_H

#if defined( _MSC_VER ) && _MSC_VER > 1000
#pragma once
#pragma warning(disable:4786)
#endif

#if defined( _MSC_VER ) && _MSC_VER >= 1400
#pragma warning (push)
#pragma warning (disable:4996)  // Disable VC8 'safe' warnings
#endif

#if defined( __BORLANDC__ )
#pragma warn -8058 // Disable "code in header" warning
#pragma warn -8008 // Disable "condition always true"
#pragma warn -8066 // Disable "unreachable code" warnings
using std::_atoi64;
#endif

// VC 6 does not place library functions in the std namespace
#if defined( _MSC_VER ) && _MSC_VER <= 1200
#define LMXSTDCRT
#else
#define LMXSTDCRT std
#endif


#if defined( __GNUC__ ) && __GNUC__ == 2
    #define __GLIBCPP_WCHAR_SUPPORT__
#endif
#include <string>
#include <cstdio>
#include <vector>
#include <list>
#include <deque>
#include <iosfwd>
#include <sstream>
#if defined( _MSC_VER ) && ! defined( UNDER_CE ) || defined( __BORLANDC__ ) || defined( __GNUC__ ) && __GNUC__ >= 3
    #include <locale>
#endif

namespace lmx {

//------------------------------------------------------------------------------
//                                Useful macros
//------------------------------------------------------------------------------

// V3.0 - We like NULL for clarity, but it's not always portable!
#ifndef LMXNULL
#define LMXNULL 0
#endif

// USER: ASSERTS - This option allows control of the handling of assertions.
// -You may use a custom assertion handler for debug code, or have different
// -handling of assertion code depending on whether the code is running in
// -debug or release mode.  For example, you may wish to throw an exception in
// -release mode. (V2.7)(v3.3 - moved position in file)

#define lmx_assert( x )     assert( x )

//------------------------------------------------------------------------------
//                     Definitions to configure operation
//------------------------------------------------------------------------------

// V2.11.1 - Allowed #defines to be optionally set on the C++ compiler command line

// USER: CTRL_DEFS - Set these #defines to tailor your code builds

#ifdef _DEBUG
    #ifndef LMX_USER_VALIDATE
        #define LMX_USER_VALIDATE
    #endif
    #ifndef LMX_USE_USER_VALIDATE_EXCEPTIONS
        #define LMX_USE_USER_VALIDATE_EXCEPTIONS 0
                // Set to 1 to throw exceptions when validating user set
                // -operations
    #endif
#endif
#ifndef LMX_USE_EXCEPTIONS
    #define LMX_USE_EXCEPTIONS 0
                // Set to 1 to throw exceptions on marshalling and unmarshalling
                // -errors
#endif
#ifndef LMX_WANT_DECIMAL_FLOAT_SET
    #define LMX_WANT_DECIMAL_FLOAT_SET 1
                // Set to 1 to include special function to convert float number to
                // -decimal number taking into account fractionDigits facet of
                // -the decimal value

#endif

// Note: WANT_DLL - Under Windows there are many ways to partition your code,
// LMX generated code, and LMX Supporting Software between your main exe and
// DLLs.  These variables allow set up your system for the more common
// arrangements.  If the arrangement you wish to use is not supported by these
// flags, you may wish to change the layout of the code in the DECL_DEFS
// section.  Note that these #defines can either be set here or in the project
// settings.
// If you are statically linking all code into a stand-alone exe, use
// LMX_WANT_DLL = 0 and LMX_WANT_GEN_IMPORT_DLL = 0.  (This is also the option
// for Linux and other OSs that do not use such constructs.)
// If you want to use the LMX Supporting Software in a DLL (or create such a
// DLL) then use LMX_WANT_DLL = 1.
// If you are producing a DLL that exports LMX generated code, also use
// LMX_WANT_DLL = 1.
// If you are producing code that accesses LMX generated code that is located
// in a separate DLL, then use LMX_WANT_GEN_IMPORT_DLL = 1.

#ifndef LMX_WANT_DLL
    #define LMX_WANT_DLL 0
                // Set to 1 if a DLL version is required.
#endif
#ifndef LMX_WANT_GEN_IMPORT_DLL
    #define LMX_WANT_GEN_IMPORT_DLL 0
                // Set to 1 if a DLL version is required.
#endif

// USER: XML_STRICT - Setting this #define to 1 forces a strict interpretation
// of XML.  Setting it to 0 allows a more liberal interpretation of XML.
// In particular, when LMX_XML_STRICT == 0 references to entities that have not
// been defined are treated as if they were regular text, rather than entity
// references, UTF-16 surrogates are allowed in UTF-8 input.
// N.B. This definition only takes effect if you have the supporting software
// source code (Professional Edition).

#ifndef LMX_XML_STRICT
    #define LMX_XML_STRICT 1
#endif

// USER: STRING_SIZE - Set this #define to 1 to use narrow (UTF-8) std::strings
// for the user interface to the methods.

#ifndef LMX_NARROW_UNICODE_STRINGS
    #define LMX_NARROW_UNICODE_STRINGS 0
#endif

// USER: VALIDATE_PATTERNS - As of version 3.6 LMX validates patterns by
// default.  For compatibility with earlier versions that did not validate
// patterns set this #define to 0.

#ifndef LMX_VALIDATE_PATTERNS
    #define LMX_VALIDATE_PATTERNS 1
#endif

//------------------------------------------------------------------------------
//                       Control DLL export requirements
//------------------------------------------------------------------------------

// USER: DECL_DEFS - See the note WANT_DLL for additional guidance on this
// section.
// v3.0 - These flags have been re-arranged to allow generated code to reside
// in the same module as the code that is using it.

#if LMX_WANT_DLL == 1 || defined( LMX_MAKE_DLL )
    #ifdef LMX_PARSE_SOURCE
    # define LMX_PDECL __declspec(dllexport)
    #else
    # define LMX_PDECL __declspec(dllimport)
    #endif
    #ifdef LMX_TYPES_SOURCE
    # define LMX_TDECL __declspec(dllexport)
    #else
    # define LMX_TDECL __declspec(dllimport)
    #endif
    #ifdef LMX_REGEX_SOURCE
    # define LMX_RDECL __declspec(dllexport)
    #else
    # define LMX_RDECL __declspec(dllimport)
    #endif
#else
    #define LMX_PDECL
    #define LMX_TDECL
    #define LMX_RDECL
#endif

#if LMX_WANT_GEN_IMPORT_DLL == 1
    # define LMX_GDECL __declspec(dllimport)
#elif LMX_WANT_DLL == 1
    # define LMX_GDECL __declspec(dllexport)
#elif defined( LMX_MAKE_DLL )
    #ifdef LMX_GEN_SOURCE
    # define LMX_GDECL __declspec(dllexport)
    #else
    # define LMX_GDECL __declspec(dllimport)
    #endif
#else
    #define LMX_GDECL
#endif

// (V2.7) Inclusion of lmxinternals.h moved to lower down in file.

//------------------------------------------------------------------------------
//                          Mapping of basic types
//------------------------------------------------------------------------------

// USER: BASIC_TYPES - Modify these typedefs according to your platform
// -requirements (currently set for MS VC++, Borland and GCC)

#if defined( _MSC_VER ) || defined( __BORLANDC__ )
    typedef __int64 tlmx_int64;
    typedef unsigned __int64 tlmx_uns64;
#elif defined( __GNUC__ )
    typedef long long tlmx_int64;
    typedef unsigned long long tlmx_uns64;
#else
    // If we want to represent 64-bit values using c_big_int, we need to do
    // -that after c_big_int is defined.  Use a #define to trigger this later.
    #define LMX_REPRESENT_INT64_USING_C_BIG_INT
#endif
typedef int tlmx_int32;
typedef unsigned int tlmx_uns32;
typedef short tlmx_int16;
typedef unsigned short tlmx_uns16;
typedef signed char tlmx_int8;
typedef unsigned char tlmx_uns8;

typedef float tlmx_float;
typedef double tlmx_double;

typedef std::string tlmx_string;

#if LMX_NARROW_UNICODE_STRINGS == 1
    typedef std::string tlmx_unicode_string;
#else
    typedef std::wstring tlmx_unicode_string;
#endif

// Typedef tlmx_wide_string with the same type as tlmx_unicode_string
// for compatibility with older versions.
typedef tlmx_unicode_string tlmx_wide_string;

// USER: INT_NATIVE_FOREIGN - Define whether each integer type is mapped
// -to a C++ native integer type (e.g. short), or mapped to a defined
// -object (e.g. c_big_int).  If the type is native, set the macro
// -result to 'native'.  Otherwise set the macro result to 'foreign'.

#define MKINT8( native, foreign ) native
#define MKUNS8( native, foreign ) native
#define MKINT16( native, foreign ) native
#define MKUNS16( native, foreign ) native
#define MKINT32( native, foreign ) native
#define MKUNS32( native, foreign ) native
#if defined( _MSC_VER ) || defined( __BORLANDC__ )
    #define MKINT64( native, foreign ) native
    #define MKUNS64( native, foreign ) native
#elif defined( __GNUC__ )
    #define MKINT64( native, foreign ) native ## LL
    #define MKUNS64( native, foreign ) native ## ULL
#else
    #define MKINT64( native, foreign ) foreign
    #define MKUNS64( native, foreign ) foreign
#endif

// USER: FULL_UNICODE_CHAR - Minimum 20 bits to hold whole range of Unicode chars
// -(0x0-0x10ffff).  Signed so that -1 can mean end-of-data.
typedef tlmx_int32 t_unicode_char;

} // End of namespace lmx

//------------------------------------------------------------------------------
//            Default version of compound types supplied with LMX
//------------------------------------------------------------------------------

#include "lmxtypes.h"

namespace lmx {

#if defined( LMX_REPRESENT_INT64_USING_C_BIG_INT )
    typedef c_big_int tlmx_int64;
    typedef c_big_int tlmx_uns64;
#endif

//------------------------------------------------------------------------------
//                       Mapping of compound types
//------------------------------------------------------------------------------

// USER: COMPOUND_TYPES - Change these typedefs as required to match your
// -requirements
typedef c_big_int       tc_big_int;
typedef c_decimal       tc_decimal;
typedef c_duration      tc_duration;
typedef c_datetime      tc_datetime;
typedef c_time          tc_time;
typedef c_date          tc_date;
typedef c_gyearmonth    tc_gyearmonth;
typedef c_gyear         tc_gyear;
typedef c_gmonthday     tc_gmonthday;
typedef c_gday          tc_gday;
typedef c_gmonth        tc_gmonth;
typedef c_binary        tc_binary;
typedef ct_qname< tlmx_unicode_string > tc_qname;  // V3.2.4


#if defined( LMX_VALIDATE_PATTERNS ) && LMX_VALIDATE_PATTERNS == 1
    typedef c_pattern       tc_pattern;
#else
    typedef c_pattern_null  tc_pattern;
#endif

//------------------------------------------------------------------------------
//                     Include lmxinternals.h file (V2.7)
//------------------------------------------------------------------------------

} // End of namespace lmx

#include "lmxinternals.h"

namespace lmx {

//------------------------------------------------------------------------------
//                         Input Converters Helpers
//------------------------------------------------------------------------------

LMX_PDECL double float_zero();  // V2.9: Now in lmxparse.cpp

inline double atofloat( const std::string & ar_value )
{
    if( ar_value == "NaN" )     // See portability note: IEEE754
        return 0.0 / float_zero();
    else if( ar_value == "INF" )
        return 1.0 / float_zero();
    else if( ar_value == "-INF" )
        return -1.0 / float_zero();
    else
        {
        // Make sure conversion is done in the "C" locale rather than any other
        // -locale that might have been set up.  (Ensures floats interpreted as
        // -1.2345 as per the standard, rather than potentially as 1,2345.)
#if defined( _MSC_VER ) && _MSC_VER >= 1300 && ! defined( UNDER_CE )
        // More efficient than the second method
        typedef std::string::value_type charT;
        typedef std::string::const_iterator iterT;
        std::basic_ios<char> l_format(0);
        l_format.imbue( std::locale::classic() );
        std::ios_base::iostate l_error;
        double l_result;
        std::use_facet< std::num_get< charT, iterT > >(
                std::locale::classic() ).get (
                        ar_value.begin(),
                        ar_value.end(),
                        l_format,
                        l_error,
                        l_result );
        return l_result;
#elif defined( _MSC_VER ) && ! defined( UNDER_CE ) || defined( __GNUC__ ) && __GNUC__ >= 3
        // For systems that don't support use_facet<>
        std::istringstream l_in_string( ar_value );
        l_in_string.imbue( std::locale::classic() );
        double l_result;
        l_in_string >> l_result;
        return l_result;
#else
        // For systems that don't support locales
        return LMXSTDCRT::atof( ar_value.c_str() );
#endif
        }
}

//------------------------------------------------------------------------------
//        Templates and template specialisations for input converters
//------------------------------------------------------------------------------

// USER: TEMPLATE_INPUT_CONVERTERS - Add new template specialisations if required

template <typename T>
inline void v_to_o( T &ar_out, const std::string &ar_in )
{
    ar_out = ar_in;
}

// Format of the conditional macros is: MKINT64( native, foreign )
#if MKINT64( 1, 0 ) == 1
    template <>
    inline void v_to_o( tlmx_int64 &ar_out, const std::string &ar_in )
    {
    #if defined( _MSC_VER ) || defined( __BORLANDC__ )
        ar_out = _atoi64( ar_in.c_str() );
    #elif defined( __GNUC__ )
        ar_out = 0;
        LMXSTDCRT::sscanf( ar_in.c_str(), "%Ld", &ar_out );
    #else
        #error Need support for reading 64bit integers
    #endif
    }
#endif

#if MKINT32( 1, 0 ) == 1
    template <>
    inline void v_to_o( tlmx_int32 &ar_out, const std::string &ar_in )
    {
        ar_out = LMXSTDCRT::atoi( ar_in.c_str() );
    }
#endif

#if MKINT16( 1, 0 ) == 1
    template <>
    inline void v_to_o( tlmx_int16 &ar_out, const std::string &ar_in )
    {
        ar_out = LMXSTDCRT::atoi( ar_in.c_str() );
    }
#endif

#if MKINT8( 1, 0 ) == 1
    template <>
    inline void v_to_o( tlmx_int8 &ar_out, const std::string &ar_in )
    {
        ar_out = LMXSTDCRT::atoi( ar_in.c_str() );
    }
#endif

#if MKUNS64( 1, 0 ) == 1
    template <>
    inline void v_to_o( tlmx_uns64 &ar_out, const std::string &ar_in )
    {
        ar_out = 0;
    #if defined( _MSC_VER ) || defined( __BORLANDC__ )
        LMXSTDCRT::sscanf( ar_in.c_str(), "%I64u", &ar_out );
    #elif defined( __GNUC__ )
        LMXSTDCRT::sscanf( ar_in.c_str(), "%Lu", &ar_out );
    #else
        #error Need support for reading 64bit integers
    #endif
    }
#endif

#if MKUNS32( 1, 0 ) == 1
    template <>
    inline void v_to_o( tlmx_uns32 &ar_out, const std::string &ar_in )
    {
        ar_out =  LMXSTDCRT::strtoul( ar_in.c_str(), LMXNULL, 10 );
    }
#endif

#if MKUNS16( 1, 0 ) == 1
    template <>
    inline void v_to_o( tlmx_uns16 &ar_out, const std::string &ar_in )
    {
        ar_out = LMXSTDCRT::atoi( ar_in.c_str() );
    }
#endif

#if MKUNS8( 1, 0 ) == 1
    template <>
    inline void v_to_o( tlmx_uns8 &ar_out, const std::string &ar_in )
    {
        ar_out = LMXSTDCRT::atoi( ar_in.c_str() );
    }
#endif

template <>
inline void v_to_o( float &ar_out, const std::string &ar_in )
{
    ar_out = static_cast<float>( atofloat( ar_in ) );
}

template <>
inline void v_to_o( double &ar_out, const std::string &ar_in )
{
    ar_out = atofloat( ar_in );
}

template <>
inline void v_to_o( bool &ar_out, const std::string &ar_in )
{
    //true, false, 1, 0
    if( ar_in == "true" || ar_in == "1" )
        ar_out = true;
    else
        ar_out = false;
}

template <>
inline void v_to_o( std::string &ar_out, const std::string &ar_in )
{
    vtostring( ar_out, ar_in );
}

template <>
inline void v_to_o( std::wstring &ar_out, const std::string &ar_in )
{
    vtostring( ar_out, ar_in );
}

inline void from_hex( c_binary &ar_out, const std::string &ar_in )
{
    ar_out.from_hex( ar_in );
}

inline void from_base64( c_binary &ar_out, const std::string &ar_in )
{
    ar_out.from_base64( ar_in );
}

//------------------------------------------------------------------------------
//                   End of template input converters
//------------------------------------------------------------------------------

// (V2.7) inittowstring prototype removed.


// USER: BASIC_OUTPUT_CONVERTERS - Adjust these to your platform requirements
// -(Currently set for MSVC++, Borland & GCC)

#if defined( _MSC_VER ) && _MSC_VER <= 1200
        // MSVC 6 ostream library does not natively support __int64

inline std::ostream & operator << ( std::ostream &ar_os, unsigned __int64 a_uns64 )
{
    char lc_buf[22];
    LMXSTDCRT::sprintf( lc_buf, "%I64u", a_uns64 );
    ar_os << lc_buf;
    return ar_os;
}

inline std::ostream & operator << ( std::ostream &ar_os, __int64 a_int64 )
{
    // MSVC 6 ostream library does not natively support __int64
    char lc_buf[22];
    LMXSTDCRT::sprintf( lc_buf, "%I64d", a_int64 );
    ar_os << lc_buf;
    return ar_os;
}
#endif

// End of conversion functions

//------------------------------------------------------------------------------
//                           Common User Base Class
//------------------------------------------------------------------------------

// USER: USER_BASE - This class can be used as the base for all generated
// -class.  Specify '-userbase' on the command line, or select
// -'Use common user base class' in the Windows interface.

class c_user_base
{
    // You may add code here
};

//------------------------------------------------------------------------------
//                           Basic error handling function
//------------------------------------------------------------------------------

// USER: ERROR_HANDLING - Redefine this function to modify error handling
inline elmx_error lmx_error( elmx_error a_error )
{
#if LMX_USE_USER_VALIDATE_EXCEPTIONS == 1
    if( a_error != ELMX_OK )    // Avoid any chance of an unreachable code warning
        throw c_lmx_exception( a_error );
#endif
    return a_error;
}

//------------------------------------------------------------------------------
//                           Output converters
//------------------------------------------------------------------------------

// IOMANIP does not seem to be supported on all platforms, so use alternative strategy
// Define on one line to avoid misc platform eol problems. Use macro to save typing!
#define ALIASED_PTR_CLASS( x ) template< class T > class c_##x { public: c_##x( const T & a ) : p(&a) {} const T* p; };
#define ALIASED_PTR_FUNC( x ) template< class T > inline const c_##x< T > x( const T &a ) { return c_##x< T >( a ); }
#define ALIASED_PTR( x ) ALIASED_PTR_CLASS( x ) ALIASED_PTR_FUNC( x )

ALIASED_PTR( as_xml )
ALIASED_PTR( as_bool_digit )    // V3.7
ALIASED_PTR( as_hex )
ALIASED_PTR( as_base64 )
ALIASED_PTR( as_escaped )
ALIASED_PTR( as_ascii )     // Primarily for user convenience


LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_hex<c_binary> &ar_v );
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_base64<c_binary> &ar_v );
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_bool_digit<bool> &ar_v );   // V3.7
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_escaped<std::string> &ar_v );
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_ascii<std::wstring> &ar_v );

// V3.2.4 - To help with QName output
class c_as_xml_string
{
private:
    std::string s;
public:
    c_as_xml_string( const std::string & a ) : s(a) {}
    c_as_xml_string( const std::wstring & a ) { convert( &s, a ); }
    const std::string &get_s() const { return s; }
};

//------------------------------------------------------------------------------
//        Templates and template specialisations for output converters
//------------------------------------------------------------------------------

// USER: TEMPLATE_OUTPUT_CONVERTERS - Add new template specialisations if required

template <typename T>
inline std::ostream & operator << ( std::ostream & ar_os, const c_as_xml<T> &ar_v )
{
    ar_os << *(ar_v.p);
    return ar_os;
}

template <>
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_xml<std::wstring> &ar_v );
template <>
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_xml<std::string> &ar_v );
template <>
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_xml<bool> &ar_v );
template <>
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_xml<float> &ar_v );
template <>
LMX_TDECL std::ostream & operator << ( std::ostream & ar_os, const c_as_xml<double> &ar_v );

// See lmxparse.h for c_as_xml<char> and c_as_xml<unsigned char> versions.

// V3.2.4 - To help with QName output
inline std::ostream & operator << ( std::ostream & ar_os, const c_as_xml_string &ar_v )
        { ar_os << as_xml( ar_v.get_s() ); return ar_os; }

} // End of namespace lmx

#if defined( _MSC_VER ) && _MSC_VER >= 1400
#pragma warning (pop)  // Restore VC8 'safe' warnings
#endif

#endif // LMXUSER_H


//------------------------------------------------------------------------------
//                   Copyright (c) 2003-2007 Codalogic Ltd.
//------------------------------------------------------------------------------
